#include "torch/csrc/jit/operator.h"

#include "torch/csrc/autograd/profiler.h"
#include "torch/csrc/jit/interned_strings.h"
#include "torch/csrc/jit/tensor_conversions.h"
#include "torch/csrc/utils/functional.h"
#include "torch/csrc/variable_tensor_functions.h"
#include "torch/csrc/autograd/generated/variable_factories.h"

#include <ATen/ATen.h>

#include <algorithm>
#include <array>
#include <cstddef>
#include <cstring>
#include <sstream>
#include <stdexcept>
#include <tuple>
#include <unordered_map>
#include <unordered_set>
#include <utility>
#include <vector>

// @generated from tools/autograd/templates/register_aten_ops.cpp

namespace torch { namespace jit {

using autograd::Variable;
using autograd::variable_list;
using at::Scalar;
using at::Tensor;
using at::IntList;
using at::TensorList;
using at::TensorOptions;
using at::DeviceGuard;

namespace {

int deviceForInputs(Stack & stack, size_t N) {
  if(N == 0)
    return -1;
  auto & t = *(stack.end() - N);
  return t.type().is_cuda() ? (int) t.get_device() : -1;
}

template<size_t N>
std::array<bool, N> as_bool_array(const std::vector<int64_t>& vec) {
  std::array<bool, N> res;
  JIT_ASSERT(vec.size() == N);
  std::copy(vec.begin(), vec.end(), res.begin());
  return res;
}

RegisterOperators reg({
Operator(
    "aten::RoiPooling2d_backward(Tensor input, Tensor rois, int pooledHeight, int pooledWidth, float spatialScale, Tensor gradOutput, Tensor argmaxes) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("RoiPooling2d_backward");
        auto pooledHeight = tensor_as<int64_t>(std::move(peek(stack, 2, 7)));
        auto pooledWidth = tensor_as<int64_t>(std::move(peek(stack, 3, 7)));
        auto spatialScale = tensor_as<double>(std::move(peek(stack, 4, 7)));
        auto result = at::RoiPooling2d_backward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), pooledHeight, pooledWidth, spatialScale, std::move(peek(stack, 5, 7)), std::move(peek(stack, 6, 7)));
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto pooledHeight = int64_t(node->i(Symbol::attr("pooledHeight")));
      auto pooledWidth = int64_t(node->i(Symbol::attr("pooledWidth")));
      auto spatialScale = double(node->f(Symbol::attr("spatialScale")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("RoiPooling2d_backward");
    
        auto result = at::RoiPooling2d_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), pooledHeight, pooledWidth, spatialScale, std::move(peek(stack, 2, 4)), std::move(peek(stack, 3, 4)));
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::RoiPooling2d_forward(Tensor input, Tensor rois, int pooledHeight, int pooledWidth, float spatialScale) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("RoiPooling2d_forward");
        auto pooledHeight = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto pooledWidth = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto spatialScale = tensor_as<double>(std::move(peek(stack, 4, 5)));
        auto result = at::RoiPooling2d_forward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), pooledHeight, pooledWidth, spatialScale);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto pooledHeight = int64_t(node->i(Symbol::attr("pooledHeight")));
      auto pooledWidth = int64_t(node->i(Symbol::attr("pooledWidth")));
      auto spatialScale = double(node->f(Symbol::attr("spatialScale")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("RoiPooling2d_forward");
    
        auto result = at::RoiPooling2d_forward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), pooledHeight, pooledWidth, spatialScale);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__and__(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__and__");
    
        auto result = at::__and__(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__and__(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__and__");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::__and__(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__and__");
    
        auto result = at::__and__(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__iand__(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__iand__");
    
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).__iand__(std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__iand__(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__iand__");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).__iand__(other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__iand__");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).__iand__(other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__ilshift__(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__ilshift__");
    
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).__ilshift__(std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__ilshift__(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__ilshift__");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).__ilshift__(other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__ilshift__");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).__ilshift__(other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__ior__(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__ior__");
    
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).__ior__(std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__ior__(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__ior__");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).__ior__(other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__ior__");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).__ior__(other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__irshift__(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__irshift__");
    
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).__irshift__(std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__irshift__(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__irshift__");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).__irshift__(other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__irshift__");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).__irshift__(other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__ixor__(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__ixor__");
    
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).__ixor__(std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__ixor__(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__ixor__");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).__ixor__(other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__ixor__");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).__ixor__(other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__lshift__(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__lshift__");
    
        auto result = at::__lshift__(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__lshift__(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__lshift__");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::__lshift__(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__lshift__");
    
        auto result = at::__lshift__(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__or__(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__or__");
    
        auto result = at::__or__(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__or__(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__or__");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::__or__(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__or__");
    
        auto result = at::__or__(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__rshift__(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__rshift__");
    
        auto result = at::__rshift__(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__rshift__(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__rshift__");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::__rshift__(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__rshift__");
    
        auto result = at::__rshift__(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__xor__(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__xor__");
    
        auto result = at::__xor__(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::__xor__(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__xor__");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::__xor__(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("__xor__");
    
        auto result = at::__xor__(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_abs(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_abs");
    
        auto result = at::_abs(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_acos(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_acos");
    
        auto result = at::_acos(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_addmv(Tensor self, Tensor mat, Tensor vec, *, Scalar beta=1, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_addmv");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 4, 5)));
        auto result = at::_addmv(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), beta, alpha);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_addmv");
    
        auto result = at::_addmv(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), beta, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_addr(Tensor self, Tensor vec1, Tensor vec2, *, Scalar beta=1, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_addr");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 4, 5)));
        auto result = at::_addr(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), beta, alpha);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_addr");
    
        auto result = at::_addr(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), beta, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_argmax(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_argmax");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_argmax(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_argmax");
    
        auto result = at::_argmax(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_argmin(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_argmin");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_argmin(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_argmin");
    
        auto result = at::_argmin(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_asin(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_asin");
    
        auto result = at::_asin(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_atan(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_atan");
    
        auto result = at::_atan(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cast_Byte(Tensor self, bool non_blocking=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Byte");
        auto non_blocking = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::_cast_Byte(std::move(peek(stack, 0, 2)), non_blocking);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto non_blocking = bool(node->i(Symbol::attr("non_blocking")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Byte");
    
        auto result = at::_cast_Byte(std::move(peek(stack, 0, 1)), non_blocking);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cast_Char(Tensor self, bool non_blocking=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Char");
        auto non_blocking = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::_cast_Char(std::move(peek(stack, 0, 2)), non_blocking);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto non_blocking = bool(node->i(Symbol::attr("non_blocking")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Char");
    
        auto result = at::_cast_Char(std::move(peek(stack, 0, 1)), non_blocking);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cast_Double(Tensor self, bool non_blocking=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Double");
        auto non_blocking = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::_cast_Double(std::move(peek(stack, 0, 2)), non_blocking);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto non_blocking = bool(node->i(Symbol::attr("non_blocking")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Double");
    
        auto result = at::_cast_Double(std::move(peek(stack, 0, 1)), non_blocking);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cast_Float(Tensor self, bool non_blocking=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Float");
        auto non_blocking = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::_cast_Float(std::move(peek(stack, 0, 2)), non_blocking);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto non_blocking = bool(node->i(Symbol::attr("non_blocking")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Float");
    
        auto result = at::_cast_Float(std::move(peek(stack, 0, 1)), non_blocking);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cast_Half(Tensor self, bool non_blocking=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Half");
        auto non_blocking = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::_cast_Half(std::move(peek(stack, 0, 2)), non_blocking);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto non_blocking = bool(node->i(Symbol::attr("non_blocking")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Half");
    
        auto result = at::_cast_Half(std::move(peek(stack, 0, 1)), non_blocking);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cast_Int(Tensor self, bool non_blocking=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Int");
        auto non_blocking = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::_cast_Int(std::move(peek(stack, 0, 2)), non_blocking);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto non_blocking = bool(node->i(Symbol::attr("non_blocking")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Int");
    
        auto result = at::_cast_Int(std::move(peek(stack, 0, 1)), non_blocking);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cast_Long(Tensor self, bool non_blocking=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Long");
        auto non_blocking = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::_cast_Long(std::move(peek(stack, 0, 2)), non_blocking);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto non_blocking = bool(node->i(Symbol::attr("non_blocking")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Long");
    
        auto result = at::_cast_Long(std::move(peek(stack, 0, 1)), non_blocking);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cast_Short(Tensor self, bool non_blocking=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Short");
        auto non_blocking = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::_cast_Short(std::move(peek(stack, 0, 2)), non_blocking);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto non_blocking = bool(node->i(Symbol::attr("non_blocking")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cast_Short");
    
        auto result = at::_cast_Short(std::move(peek(stack, 0, 1)), non_blocking);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cat(Tensor[] tensors, int dim=0) -> Tensor",
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cat");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 0, 1)));
        auto result = at::_cat(peekSlice(stack, 0, varargs_length - 1, varargs_length), dim);
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cat");
    
        auto result = at::_cat(peekSlice(stack, 0, varargs_length - 0, varargs_length), dim);
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_ceil(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_ceil");
    
        auto result = at::_ceil(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_convolution(Tensor input, Tensor weight, Tensor? bias, int[] stride, int[] padding, int[] dilation, bool transposed, int[] output_padding, int groups, bool benchmark, bool deterministic, bool cudnn_enabled) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_convolution");
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 12)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 12)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 12)));
        auto transposed = tensor_as<bool>(std::move(peek(stack, 6, 12)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 7, 12)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 8, 12)));
        auto benchmark = tensor_as<bool>(std::move(peek(stack, 9, 12)));
        auto deterministic = tensor_as<bool>(std::move(peek(stack, 10, 12)));
        auto cudnn_enabled = tensor_as<bool>(std::move(peek(stack, 11, 12)));
        auto result = at::_convolution(std::move(peek(stack, 0, 12)), std::move(peek(stack, 1, 12)), std::move(peek(stack, 2, 12)), stride, padding, dilation, transposed, output_padding, groups, benchmark, deterministic, cudnn_enabled);
        drop(stack, 12);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto transposed = bool(node->i(Symbol::attr("transposed")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto benchmark = bool(node->i(Symbol::attr("benchmark")));
      auto deterministic = bool(node->i(Symbol::attr("deterministic")));
      auto cudnn_enabled = bool(node->i(Symbol::attr("cudnn_enabled")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_convolution");
    
        auto result = at::_convolution(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), stride, padding, dilation, transposed, output_padding, groups, benchmark, deterministic, cudnn_enabled);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_convolution_double_backward(Tensor? ggI, Tensor? ggW, Tensor? ggb, Tensor gO, Tensor weight, Tensor self, int[] stride, int[] padding, int[] dilation, bool transposed, int[] output_padding, int groups, bool benchmark, bool deterministic, bool cudnn_enabled, bool[3] output_mask) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_convolution_double_backward");
        auto stride = tensor_as<IntList>(std::move(peek(stack, 6, 16)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 7, 16)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 8, 16)));
        auto transposed = tensor_as<bool>(std::move(peek(stack, 9, 16)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 10, 16)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 11, 16)));
        auto benchmark = tensor_as<bool>(std::move(peek(stack, 12, 16)));
        auto deterministic = tensor_as<bool>(std::move(peek(stack, 13, 16)));
        auto cudnn_enabled = tensor_as<bool>(std::move(peek(stack, 14, 16)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 15, 16)));
        auto result = at::_convolution_double_backward(std::move(peek(stack, 0, 16)), std::move(peek(stack, 1, 16)), std::move(peek(stack, 2, 16)), std::move(peek(stack, 3, 16)), std::move(peek(stack, 4, 16)), std::move(peek(stack, 5, 16)), stride, padding, dilation, transposed, output_padding, groups, benchmark, deterministic, cudnn_enabled, output_mask);
        drop(stack, 16);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto transposed = bool(node->i(Symbol::attr("transposed")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto benchmark = bool(node->i(Symbol::attr("benchmark")));
      auto deterministic = bool(node->i(Symbol::attr("deterministic")));
      auto cudnn_enabled = bool(node->i(Symbol::attr("cudnn_enabled")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_convolution_double_backward");
    
        auto result = at::_convolution_double_backward(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), std::move(peek(stack, 2, 6)), std::move(peek(stack, 3, 6)), std::move(peek(stack, 4, 6)), std::move(peek(stack, 5, 6)), stride, padding, dilation, transposed, output_padding, groups, benchmark, deterministic, cudnn_enabled, output_mask);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_convolution_nogroup(Tensor input, Tensor weight, Tensor? bias, int[] stride, int[] padding, int[] dilation, bool transposed, int[] output_padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_convolution_nogroup");
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto transposed = tensor_as<bool>(std::move(peek(stack, 6, 8)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 7, 8)));
        auto result = at::_convolution_nogroup(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), stride, padding, dilation, transposed, output_padding);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto transposed = bool(node->i(Symbol::attr("transposed")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_convolution_nogroup");
    
        auto result = at::_convolution_nogroup(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), stride, padding, dilation, transposed, output_padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cos(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cos");
    
        auto result = at::_cos(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cosh(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cosh");
    
        auto result = at::_cosh(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cudnn_rnn(Tensor input, Tensor[] weight, int weight_stride0, Tensor? weight_buf, Tensor hx, Tensor? cx, int mode, int hidden_size, int num_layers, bool batch_first, float dropout, bool train, bool bidirectional, int[] batch_sizes, Tensor? dropout_state) -> (Tensor, Tensor, Tensor, Tensor, Tensor)",
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cudnn_rnn");
        auto weight_stride0 = tensor_as<int64_t>(std::move(peek(stack, 1, 14)));
        auto mode = tensor_as<int64_t>(std::move(peek(stack, 5, 14)));
        auto hidden_size = tensor_as<int64_t>(std::move(peek(stack, 6, 14)));
        auto num_layers = tensor_as<int64_t>(std::move(peek(stack, 7, 14)));
        auto batch_first = tensor_as<bool>(std::move(peek(stack, 8, 14)));
        auto dropout = tensor_as<double>(std::move(peek(stack, 9, 14)));
        auto train = tensor_as<bool>(std::move(peek(stack, 10, 14)));
        auto bidirectional = tensor_as<bool>(std::move(peek(stack, 11, 14)));
        auto batch_sizes = tensor_as<IntList>(std::move(peek(stack, 12, 14)));
        auto result = at::_cudnn_rnn(std::move(peek(stack, 0, varargs_length)), peekSlice(stack, 1, varargs_length - 14, varargs_length), weight_stride0, std::move(peek(stack, 2, 14)), std::move(peek(stack, 3, 14)), std::move(peek(stack, 4, 14)), mode, hidden_size, num_layers, batch_first, dropout, train, bidirectional, batch_sizes, std::move(peek(stack, 13, 14)));
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      auto weight_stride0 = int64_t(node->i(Symbol::attr("weight_stride0")));
      auto mode = int64_t(node->i(Symbol::attr("mode")));
      auto hidden_size = int64_t(node->i(Symbol::attr("hidden_size")));
      auto num_layers = int64_t(node->i(Symbol::attr("num_layers")));
      auto batch_first = bool(node->i(Symbol::attr("batch_first")));
      auto dropout = double(node->f(Symbol::attr("dropout")));
      auto train = bool(node->i(Symbol::attr("train")));
      auto bidirectional = bool(node->i(Symbol::attr("bidirectional")));
      auto batch_sizes = std::vector<int64_t>(node->is(Symbol::attr("batch_sizes")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cudnn_rnn");
    
        auto result = at::_cudnn_rnn(std::move(peek(stack, 0, varargs_length)), peekSlice(stack, 1, varargs_length - 5, varargs_length), weight_stride0, std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), std::move(peek(stack, 3, 5)), mode, hidden_size, num_layers, batch_first, dropout, train, bidirectional, batch_sizes, std::move(peek(stack, 4, 5)));
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cudnn_rnn_backward(Tensor input, Tensor[] weight, int weight_stride0, Tensor weight_buf, Tensor hx, Tensor? cx, Tensor output, Tensor? grad_output, Tensor? grad_hy, Tensor? grad_cy, int mode, int hidden_size, int num_layers, bool batch_first, float dropout, bool train, bool bidirectional, int[] batch_sizes, Tensor? dropout_state, Tensor reserve, bool[4] output_mask) -> (Tensor, Tensor, Tensor, Tensor[])",
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cudnn_rnn_backward");
        auto weight_stride0 = tensor_as<int64_t>(std::move(peek(stack, 1, 20)));
        auto mode = tensor_as<int64_t>(std::move(peek(stack, 9, 20)));
        auto hidden_size = tensor_as<int64_t>(std::move(peek(stack, 10, 20)));
        auto num_layers = tensor_as<int64_t>(std::move(peek(stack, 11, 20)));
        auto batch_first = tensor_as<bool>(std::move(peek(stack, 12, 20)));
        auto dropout = tensor_as<double>(std::move(peek(stack, 13, 20)));
        auto train = tensor_as<bool>(std::move(peek(stack, 14, 20)));
        auto bidirectional = tensor_as<bool>(std::move(peek(stack, 15, 20)));
        auto batch_sizes = tensor_as<IntList>(std::move(peek(stack, 16, 20)));
        auto output_mask = tensor_as<std::array<bool,4>>(std::move(peek(stack, 19, 20)));
        auto result = at::_cudnn_rnn_backward(std::move(peek(stack, 0, varargs_length)), peekSlice(stack, 1, varargs_length - 20, varargs_length), weight_stride0, std::move(peek(stack, 2, 20)), std::move(peek(stack, 3, 20)), std::move(peek(stack, 4, 20)), std::move(peek(stack, 5, 20)), std::move(peek(stack, 6, 20)), std::move(peek(stack, 7, 20)), std::move(peek(stack, 8, 20)), mode, hidden_size, num_layers, batch_first, dropout, train, bidirectional, batch_sizes, std::move(peek(stack, 17, 20)), std::move(peek(stack, 18, 20)), output_mask);
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      auto weight_stride0 = int64_t(node->i(Symbol::attr("weight_stride0")));
      auto mode = int64_t(node->i(Symbol::attr("mode")));
      auto hidden_size = int64_t(node->i(Symbol::attr("hidden_size")));
      auto num_layers = int64_t(node->i(Symbol::attr("num_layers")));
      auto batch_first = bool(node->i(Symbol::attr("batch_first")));
      auto dropout = double(node->f(Symbol::attr("dropout")));
      auto train = bool(node->i(Symbol::attr("train")));
      auto bidirectional = bool(node->i(Symbol::attr("bidirectional")));
      auto batch_sizes = std::vector<int64_t>(node->is(Symbol::attr("batch_sizes")));
      auto output_mask = as_bool_array<4>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cudnn_rnn_backward");
    
        auto result = at::_cudnn_rnn_backward(std::move(peek(stack, 0, varargs_length)), peekSlice(stack, 1, varargs_length - 10, varargs_length), weight_stride0, std::move(peek(stack, 1, 10)), std::move(peek(stack, 2, 10)), std::move(peek(stack, 3, 10)), std::move(peek(stack, 4, 10)), std::move(peek(stack, 5, 10)), std::move(peek(stack, 6, 10)), std::move(peek(stack, 7, 10)), mode, hidden_size, num_layers, batch_first, dropout, train, bidirectional, batch_sizes, std::move(peek(stack, 8, 10)), std::move(peek(stack, 9, 10)), output_mask);
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cudnn_rnn_flatten_weight(Tensor[] weight_arr, int weight_stride0, int input_size, int mode, int hidden_size, int num_layers, bool batch_first, bool bidirectional) -> Tensor",
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cudnn_rnn_flatten_weight");
        auto weight_stride0 = tensor_as<int64_t>(std::move(peek(stack, 0, 7)));
        auto input_size = tensor_as<int64_t>(std::move(peek(stack, 1, 7)));
        auto mode = tensor_as<int64_t>(std::move(peek(stack, 2, 7)));
        auto hidden_size = tensor_as<int64_t>(std::move(peek(stack, 3, 7)));
        auto num_layers = tensor_as<int64_t>(std::move(peek(stack, 4, 7)));
        auto batch_first = tensor_as<bool>(std::move(peek(stack, 5, 7)));
        auto bidirectional = tensor_as<bool>(std::move(peek(stack, 6, 7)));
        auto result = at::_cudnn_rnn_flatten_weight(peekSlice(stack, 0, varargs_length - 7, varargs_length), weight_stride0, input_size, mode, hidden_size, num_layers, batch_first, bidirectional);
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      auto weight_stride0 = int64_t(node->i(Symbol::attr("weight_stride0")));
      auto input_size = int64_t(node->i(Symbol::attr("input_size")));
      auto mode = int64_t(node->i(Symbol::attr("mode")));
      auto hidden_size = int64_t(node->i(Symbol::attr("hidden_size")));
      auto num_layers = int64_t(node->i(Symbol::attr("num_layers")));
      auto batch_first = bool(node->i(Symbol::attr("batch_first")));
      auto bidirectional = bool(node->i(Symbol::attr("bidirectional")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cudnn_rnn_flatten_weight");
    
        auto result = at::_cudnn_rnn_flatten_weight(peekSlice(stack, 0, varargs_length - 0, varargs_length), weight_stride0, input_size, mode, hidden_size, num_layers, batch_first, bidirectional);
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cufft_get_plan_cache_max_size() -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cufft_get_plan_cache_max_size");
    
        auto result = at::_cufft_get_plan_cache_max_size();
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cufft_get_plan_cache_size() -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cufft_get_plan_cache_size");
    
        auto result = at::_cufft_get_plan_cache_size();
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cumprod(Tensor self, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cumprod");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::_cumprod(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cumprod");
    
        auto result = at::_cumprod(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_cumsum(Tensor self, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cumsum");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::_cumsum(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_cumsum");
    
        auto result = at::_cumsum(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_denseDims(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_denseDims");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1)))._denseDims();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_dimI(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_dimI");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1)))._dimI();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_dimV(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_dimV");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1)))._dimV();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_dim_arange(Tensor like, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_dim_arange");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::_dim_arange(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_dim_arange");
    
        auto result = at::_dim_arange(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_dirichlet_grad(Tensor x, Tensor alpha, Tensor total) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_dirichlet_grad");
    
        auto result = at::_dirichlet_grad(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_dot(Tensor self, Tensor tensor) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_dot");
    
        auto result = at::_dot(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_embedding_bag(Tensor weight, Tensor indices, Tensor offsets, bool scale_grad_by_freq=False, int mode=0, bool sparse=False) -> (Tensor, Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_embedding_bag");
        auto scale_grad_by_freq = tensor_as<bool>(std::move(peek(stack, 3, 6)));
        auto mode = tensor_as<int64_t>(std::move(peek(stack, 4, 6)));
        auto sparse = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::_embedding_bag(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), std::move(peek(stack, 2, 6)), scale_grad_by_freq, mode, sparse);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto scale_grad_by_freq = bool(node->i(Symbol::attr("scale_grad_by_freq")));
      auto mode = int64_t(node->i(Symbol::attr("mode")));
      auto sparse = bool(node->i(Symbol::attr("sparse")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_embedding_bag");
    
        auto result = at::_embedding_bag(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), scale_grad_by_freq, mode, sparse);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_embedding_bag_backward(Tensor grad, Tensor indices, Tensor offsets, Tensor offset2bag, Tensor bag_size, Tensor maximum_indices, int num_weights, bool scale_grad_by_freq, int mode, bool sparse) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_embedding_bag_backward");
        auto num_weights = tensor_as<int64_t>(std::move(peek(stack, 6, 10)));
        auto scale_grad_by_freq = tensor_as<bool>(std::move(peek(stack, 7, 10)));
        auto mode = tensor_as<int64_t>(std::move(peek(stack, 8, 10)));
        auto sparse = tensor_as<bool>(std::move(peek(stack, 9, 10)));
        auto result = at::_embedding_bag_backward(std::move(peek(stack, 0, 10)), std::move(peek(stack, 1, 10)), std::move(peek(stack, 2, 10)), std::move(peek(stack, 3, 10)), std::move(peek(stack, 4, 10)), std::move(peek(stack, 5, 10)), num_weights, scale_grad_by_freq, mode, sparse);
        drop(stack, 10);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto num_weights = int64_t(node->i(Symbol::attr("num_weights")));
      auto scale_grad_by_freq = bool(node->i(Symbol::attr("scale_grad_by_freq")));
      auto mode = int64_t(node->i(Symbol::attr("mode")));
      auto sparse = bool(node->i(Symbol::attr("sparse")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_embedding_bag_backward");
    
        auto result = at::_embedding_bag_backward(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), std::move(peek(stack, 2, 6)), std::move(peek(stack, 3, 6)), std::move(peek(stack, 4, 6)), std::move(peek(stack, 5, 6)), num_weights, scale_grad_by_freq, mode, sparse);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_embedding_bag_dense_backward(Tensor grad, Tensor indices, Tensor offsets, Tensor offset2bag, Tensor bag_size, Tensor maximum_indices, int num_weights, bool scale_grad_by_freq, int mode) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_embedding_bag_dense_backward");
        auto num_weights = tensor_as<int64_t>(std::move(peek(stack, 6, 9)));
        auto scale_grad_by_freq = tensor_as<bool>(std::move(peek(stack, 7, 9)));
        auto mode = tensor_as<int64_t>(std::move(peek(stack, 8, 9)));
        auto result = at::_embedding_bag_dense_backward(std::move(peek(stack, 0, 9)), std::move(peek(stack, 1, 9)), std::move(peek(stack, 2, 9)), std::move(peek(stack, 3, 9)), std::move(peek(stack, 4, 9)), std::move(peek(stack, 5, 9)), num_weights, scale_grad_by_freq, mode);
        drop(stack, 9);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto num_weights = int64_t(node->i(Symbol::attr("num_weights")));
      auto scale_grad_by_freq = bool(node->i(Symbol::attr("scale_grad_by_freq")));
      auto mode = int64_t(node->i(Symbol::attr("mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_embedding_bag_dense_backward");
    
        auto result = at::_embedding_bag_dense_backward(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), std::move(peek(stack, 2, 6)), std::move(peek(stack, 3, 6)), std::move(peek(stack, 4, 6)), std::move(peek(stack, 5, 6)), num_weights, scale_grad_by_freq, mode);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_embedding_bag_sparse_backward(Tensor grad, Tensor indices, Tensor offsets, Tensor offset2bag, Tensor bag_size, int num_weights, bool scale_grad_by_freq, int mode) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_embedding_bag_sparse_backward");
        auto num_weights = tensor_as<int64_t>(std::move(peek(stack, 5, 8)));
        auto scale_grad_by_freq = tensor_as<bool>(std::move(peek(stack, 6, 8)));
        auto mode = tensor_as<int64_t>(std::move(peek(stack, 7, 8)));
        auto result = at::_embedding_bag_sparse_backward(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), std::move(peek(stack, 3, 8)), std::move(peek(stack, 4, 8)), num_weights, scale_grad_by_freq, mode);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto num_weights = int64_t(node->i(Symbol::attr("num_weights")));
      auto scale_grad_by_freq = bool(node->i(Symbol::attr("scale_grad_by_freq")));
      auto mode = int64_t(node->i(Symbol::attr("mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_embedding_bag_sparse_backward");
    
        auto result = at::_embedding_bag_sparse_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), num_weights, scale_grad_by_freq, mode);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_erf(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_erf");
    
        auto result = at::_erf(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_exp(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_exp");
    
        auto result = at::_exp(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_expm1(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_expm1");
    
        auto result = at::_expm1(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_fft_with_size(Tensor self, int signal_ndim, bool complex_input, bool complex_output, bool inverse, int[] checked_signal_sizes, bool normalized, bool onesided, int[] output_sizes) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_fft_with_size");
        auto signal_ndim = tensor_as<int64_t>(std::move(peek(stack, 1, 9)));
        auto complex_input = tensor_as<bool>(std::move(peek(stack, 2, 9)));
        auto complex_output = tensor_as<bool>(std::move(peek(stack, 3, 9)));
        auto inverse = tensor_as<bool>(std::move(peek(stack, 4, 9)));
        auto checked_signal_sizes = tensor_as<IntList>(std::move(peek(stack, 5, 9)));
        auto normalized = tensor_as<bool>(std::move(peek(stack, 6, 9)));
        auto onesided = tensor_as<bool>(std::move(peek(stack, 7, 9)));
        auto output_sizes = tensor_as<IntList>(std::move(peek(stack, 8, 9)));
        auto result = at::_fft_with_size(std::move(peek(stack, 0, 9)), signal_ndim, complex_input, complex_output, inverse, checked_signal_sizes, normalized, onesided, output_sizes);
        drop(stack, 9);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto signal_ndim = int64_t(node->i(Symbol::attr("signal_ndim")));
      auto complex_input = bool(node->i(Symbol::attr("complex_input")));
      auto complex_output = bool(node->i(Symbol::attr("complex_output")));
      auto inverse = bool(node->i(Symbol::attr("inverse")));
      auto checked_signal_sizes = std::vector<int64_t>(node->is(Symbol::attr("checked_signal_sizes")));
      auto normalized = bool(node->i(Symbol::attr("normalized")));
      auto onesided = bool(node->i(Symbol::attr("onesided")));
      auto output_sizes = std::vector<int64_t>(node->is(Symbol::attr("output_sizes")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_fft_with_size");
    
        auto result = at::_fft_with_size(std::move(peek(stack, 0, 1)), signal_ndim, complex_input, complex_output, inverse, checked_signal_sizes, normalized, onesided, output_sizes);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_floor(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_floor");
    
        auto result = at::_floor(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_ger(Tensor self, Tensor vec2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_ger");
    
        auto result = at::_ger(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_gesv_helper(Tensor self, Tensor A) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_gesv_helper");
    
        auto result = at::_gesv_helper(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_gesv_single(Tensor self, Tensor A) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_gesv_single");
    
        auto result = at::_gesv_single(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_indices(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_indices");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1)))._indices();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_log(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_log");
    
        auto result = at::_log(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_log10(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_log10");
    
        auto result = at::_log10(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_log1p(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_log1p");
    
        auto result = at::_log1p(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_log2(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_log2");
    
        auto result = at::_log2(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_mm(Tensor self, Tensor mat2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_mm");
    
        auto result = at::_mm(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_mv(Tensor self, Tensor vec) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_mv");
    
        auto result = at::_mv(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_nnz(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_nnz");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1)))._nnz();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_prod(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_prod");
    
        auto result = at::_prod(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_prod(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_prod");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_prod(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_prod");
    
        auto result = at::_prod(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_prodall(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_prodall");
    
        auto result = at::_prodall(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_round(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_round");
    
        auto result = at::_round(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_rsqrt(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_rsqrt");
    
        auto result = at::_rsqrt(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_s_where(Tensor condition, Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_s_where");
    
        auto result = at::_s_where(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sigmoid(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sigmoid");
    
        auto result = at::_sigmoid(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sigmoid_backward(Tensor grad_output, Tensor output) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sigmoid_backward");
    
        auto result = at::_sigmoid_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sigmoid_forward(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sigmoid_forward");
    
        auto result = at::_sigmoid_forward(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sin(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sin");
    
        auto result = at::_sin(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sinh(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sinh");
    
        auto result = at::_sinh(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sparseDims(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sparseDims");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1)))._sparseDims();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sparse_coo_tensor_unsafe(Tensor indices, Tensor values, int[] size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sparse_coo_tensor_unsafe");
        auto size = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::_sparse_coo_tensor_unsafe(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), size);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sparse_coo_tensor_unsafe");
    
        auto result = at::_sparse_coo_tensor_unsafe(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sqrt(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sqrt");
    
        auto result = at::_sqrt(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_standard_gamma(Tensor self, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_standard_gamma");
    
        auto result = at::_standard_gamma(std::move(peek(stack, 0, 1)), nullptr);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_standard_gamma_grad(Tensor self, Tensor output) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_standard_gamma_grad");
    
        auto result = at::_standard_gamma_grad(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sum(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sum");
    
        auto result = at::_sum(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sum(Tensor self, int[1] dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sum");
        auto dim = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_sum(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = std::vector<int64_t>(node->is(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sum");
    
        auto result = at::_sum(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_sumall(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_sumall");
    
        auto result = at::_sumall(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_tan(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_tan");
    
        auto result = at::_tan(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_tanh(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_tanh");
    
        auto result = at::_tanh(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_tanh_backward(Tensor grad_output, Tensor output) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_tanh_backward");
    
        auto result = at::_tanh_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_tanh_forward(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_tanh_forward");
    
        auto result = at::_tanh_forward(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_all(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_all");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_th_all(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_all");
    
        auto result = at::_th_all(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_any(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_any");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_th_any(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_any");
    
        auto result = at::_th_any(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_bernoulli(Tensor self, *, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_bernoulli");
    
        auto result = at::_th_bernoulli(std::move(peek(stack, 0, 1)), nullptr);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_get_device(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_get_device");
    
        auto result = at::_th_get_device(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_kthvalue(Tensor self, int k, int dim=-1, bool keepdim=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_kthvalue");
        auto k = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::_th_kthvalue(std::move(peek(stack, 0, 4)), k, dim, keepdim);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto k = int64_t(node->i(Symbol::attr("k")));
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_kthvalue");
    
        auto result = at::_th_kthvalue(std::move(peek(stack, 0, 1)), k, dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_max(Tensor self, int dim, bool keepdim=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_max");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_th_max(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_max");
    
        auto result = at::_th_max(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_median(Tensor self, int dim, bool keepdim=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_median");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_th_median(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_median");
    
        auto result = at::_th_median(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_min(Tensor self, int dim, bool keepdim=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_min");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_th_min(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_min");
    
        auto result = at::_th_min(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_mode(Tensor self, int dim=-1, bool keepdim=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_mode");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_th_mode(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_mode");
    
        auto result = at::_th_mode(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_norm(Tensor self, Scalar p, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_norm");
        auto p = tensor_as<Scalar>(std::move(peek(stack, 1, 4)));
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::_th_norm(std::move(peek(stack, 0, 4)), p, dim, keepdim);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = Scalar(node->t(Symbol::attr("p")));
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_norm");
    
        auto result = at::_th_norm(std::move(peek(stack, 0, 1)), p, dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_prod(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_prod");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_th_prod(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_prod");
    
        auto result = at::_th_prod(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_std(Tensor self, int dim, bool unbiased=True, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_std");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto unbiased = tensor_as<bool>(std::move(peek(stack, 2, 4)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::_th_std(std::move(peek(stack, 0, 4)), dim, unbiased, keepdim);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto unbiased = bool(node->i(Symbol::attr("unbiased")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_std");
    
        auto result = at::_th_std(std::move(peek(stack, 0, 1)), dim, unbiased, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_sum(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_sum");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_th_sum(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_sum");
    
        auto result = at::_th_sum(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_tanh(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_tanh");
    
        auto result = at::_th_tanh(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_th_var(Tensor self, int dim, bool unbiased=True, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_var");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto unbiased = tensor_as<bool>(std::move(peek(stack, 2, 4)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::_th_var(std::move(peek(stack, 0, 4)), dim, unbiased, keepdim);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto unbiased = bool(node->i(Symbol::attr("unbiased")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_th_var");
    
        auto result = at::_th_var(std::move(peek(stack, 0, 1)), dim, unbiased, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_trilinear(Tensor i1, Tensor i2, Tensor i3, int[] expand1, int[] expand2, int[] expand3, int[] sumdim, int unroll_dim=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_trilinear");
        auto expand1 = tensor_as<IntList>(std::move(peek(stack, 3, 8)));
        auto expand2 = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto expand3 = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto sumdim = tensor_as<IntList>(std::move(peek(stack, 6, 8)));
        auto unroll_dim = tensor_as<int64_t>(std::move(peek(stack, 7, 8)));
        auto result = at::_trilinear(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), expand1, expand2, expand3, sumdim, unroll_dim);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto expand1 = std::vector<int64_t>(node->is(Symbol::attr("expand1")));
      auto expand2 = std::vector<int64_t>(node->is(Symbol::attr("expand2")));
      auto expand3 = std::vector<int64_t>(node->is(Symbol::attr("expand3")));
      auto sumdim = std::vector<int64_t>(node->is(Symbol::attr("sumdim")));
      auto unroll_dim = int64_t(node->i(Symbol::attr("unroll_dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_trilinear");
    
        auto result = at::_trilinear(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), expand1, expand2, expand3, sumdim, unroll_dim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_trunc(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_trunc");
    
        auto result = at::_trunc(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_unique(Tensor self, bool sorted=False, bool return_inverse=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_unique");
        auto sorted = tensor_as<bool>(std::move(peek(stack, 1, 3)));
        auto return_inverse = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::_unique(std::move(peek(stack, 0, 3)), sorted, return_inverse);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto sorted = bool(node->i(Symbol::attr("sorted")));
      auto return_inverse = bool(node->i(Symbol::attr("return_inverse")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_unique");
    
        auto result = at::_unique(std::move(peek(stack, 0, 1)), sorted, return_inverse);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_unsafe_view(Tensor self, int[] size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_unsafe_view");
        auto size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::_unsafe_view(std::move(peek(stack, 0, 2)), size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_unsafe_view");
    
        auto result = at::_unsafe_view(std::move(peek(stack, 0, 1)), size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::_values(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("_values");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1)))._values();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::abs(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("abs");
    
        auto result = at::abs(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::acos(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("acos");
    
        auto result = at::acos(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_avg_pool1d(Tensor self, int[1] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool1d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::adaptive_avg_pool1d(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool1d");
    
        auto result = at::adaptive_avg_pool1d(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_avg_pool2d(Tensor self, int[2] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool2d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::adaptive_avg_pool2d(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool2d");
    
        auto result = at::adaptive_avg_pool2d(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_avg_pool2d_backward(Tensor grad_output, Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool2d_backward");
    
        auto result = at::adaptive_avg_pool2d_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_avg_pool2d_forward(Tensor self, int[2] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool2d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::adaptive_avg_pool2d_forward(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool2d_forward");
    
        auto result = at::adaptive_avg_pool2d_forward(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_avg_pool3d(Tensor self, int[3] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool3d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::adaptive_avg_pool3d(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool3d");
    
        auto result = at::adaptive_avg_pool3d(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_avg_pool3d_backward(Tensor grad_output, Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool3d_backward");
    
        auto result = at::adaptive_avg_pool3d_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_avg_pool3d_forward(Tensor self, int[3] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool3d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::adaptive_avg_pool3d_forward(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_avg_pool3d_forward");
    
        auto result = at::adaptive_avg_pool3d_forward(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_max_pool1d(Tensor self, int[1] output_size) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool1d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::adaptive_max_pool1d(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool1d");
    
        auto result = at::adaptive_max_pool1d(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_max_pool2d(Tensor self, int[2] output_size) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool2d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::adaptive_max_pool2d(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool2d");
    
        auto result = at::adaptive_max_pool2d(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_max_pool2d_backward(Tensor grad_output, Tensor self, Tensor indices) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool2d_backward");
    
        auto result = at::adaptive_max_pool2d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_max_pool2d_forward(Tensor self, int[2] output_size) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool2d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::adaptive_max_pool2d_forward(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool2d_forward");
    
        auto result = at::adaptive_max_pool2d_forward(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_max_pool3d(Tensor self, int[3] output_size) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool3d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::adaptive_max_pool3d(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool3d");
    
        auto result = at::adaptive_max_pool3d(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_max_pool3d_backward(Tensor grad_output, Tensor self, Tensor indices) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool3d_backward");
    
        auto result = at::adaptive_max_pool3d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::adaptive_max_pool3d_forward(Tensor self, int[3] output_size) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool3d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::adaptive_max_pool3d_forward(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("adaptive_max_pool3d_forward");
    
        auto result = at::adaptive_max_pool3d_forward(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::add(Tensor self, Tensor other, *, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("add");
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::add(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("add");
    
        auto result = at::add(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), alpha);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::add(Tensor self, Scalar other, *, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("add");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::add(std::move(peek(stack, 0, 3)), other, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("add");
    
        auto result = at::add(std::move(peek(stack, 0, 1)), other, alpha);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::addbmm(Tensor self, Tensor batch1, Tensor batch2, *, Scalar beta=1, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addbmm");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 4, 5)));
        auto result = at::addbmm(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), beta, alpha);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addbmm");
    
        auto result = at::addbmm(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), beta, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::addcdiv(Tensor self, Tensor tensor1, Tensor tensor2, *, Scalar value=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addcdiv");
        auto value = tensor_as<Scalar>(std::move(peek(stack, 3, 4)));
        auto result = at::addcdiv(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), value);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto value = Scalar(node->t(Symbol::attr("value")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addcdiv");
    
        auto result = at::addcdiv(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), value);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::addcmul(Tensor self, Tensor tensor1, Tensor tensor2, *, Scalar value=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addcmul");
        auto value = tensor_as<Scalar>(std::move(peek(stack, 3, 4)));
        auto result = at::addcmul(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), value);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto value = Scalar(node->t(Symbol::attr("value")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addcmul");
    
        auto result = at::addcmul(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), value);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::addmm(Tensor self, Tensor mat1, Tensor mat2, *, Scalar beta=1, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addmm");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 4, 5)));
        auto result = at::addmm(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), beta, alpha);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addmm");
    
        auto result = at::addmm(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), beta, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::addmv(Tensor self, Tensor mat, Tensor vec, *, Scalar beta=1, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addmv");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 4, 5)));
        auto result = at::addmv(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), beta, alpha);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addmv");
    
        auto result = at::addmv(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), beta, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::addr(Tensor self, Tensor vec1, Tensor vec2, *, Scalar beta=1, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addr");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 4, 5)));
        auto result = at::addr(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), beta, alpha);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("addr");
    
        auto result = at::addr(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), beta, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::alias(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("alias");
    
        auto result = at::alias(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::all(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("all");
    
        auto result = at::all(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::all(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("all");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::all(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("all");
    
        auto result = at::all(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::allclose(Tensor self, Tensor other, float rtol=1e-05, float atol=1e-08, bool equal_nan=False) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("allclose");
        auto rtol = tensor_as<double>(std::move(peek(stack, 2, 5)));
        auto atol = tensor_as<double>(std::move(peek(stack, 3, 5)));
        auto equal_nan = tensor_as<bool>(std::move(peek(stack, 4, 5)));
        auto result = at::allclose(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), rtol, atol, equal_nan);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto rtol = double(node->f(Symbol::attr("rtol")));
      auto atol = double(node->f(Symbol::attr("atol")));
      auto equal_nan = bool(node->i(Symbol::attr("equal_nan")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("allclose");
    
        auto result = at::allclose(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), rtol, atol, equal_nan);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::any(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("any");
    
        auto result = at::any(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::any(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("any");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::any(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("any");
    
        auto result = at::any(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::arange(Scalar end, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("arange");
        auto end = tensor_as<Scalar>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::arange(end, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto end = Scalar(node->t(Symbol::attr("end")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("arange");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::arange(end, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::arange(Scalar start, Scalar end, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("arange");
        auto start = tensor_as<Scalar>(std::move(peek(stack, 0, 5)));
        auto end = tensor_as<Scalar>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::arange(start, end, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto start = Scalar(node->t(Symbol::attr("start")));
      auto end = Scalar(node->t(Symbol::attr("end")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("arange");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::arange(start, end, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::arange(Scalar start, Scalar end, Scalar step, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("arange");
        auto start = tensor_as<Scalar>(std::move(peek(stack, 0, 6)));
        auto end = tensor_as<Scalar>(std::move(peek(stack, 1, 6)));
        auto step = tensor_as<Scalar>(std::move(peek(stack, 2, 6)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 3, 6)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 4, 6)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::arange(start, end, step, options);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto start = Scalar(node->t(Symbol::attr("start")));
      auto end = Scalar(node->t(Symbol::attr("end")));
      auto step = Scalar(node->t(Symbol::attr("step")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("arange");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::arange(start, end, step, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::argmax(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("argmax");
    
        auto result = at::argmax(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::argmax(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("argmax");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::argmax(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("argmax");
    
        auto result = at::argmax(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::argmin(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("argmin");
    
        auto result = at::argmin(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::argmin(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("argmin");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::argmin(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("argmin");
    
        auto result = at::argmin(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::as_strided(Tensor self, int[] size, int[] stride) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("as_strided");
        auto size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::as_strided(std::move(peek(stack, 0, 3)), size, stride);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("as_strided");
    
        auto result = at::as_strided(std::move(peek(stack, 0, 1)), size, stride);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::as_strided(Tensor self, int[] size, int[] stride, int storage_offset) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("as_strided");
        auto size = tensor_as<IntList>(std::move(peek(stack, 1, 4)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 4)));
        auto storage_offset = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::as_strided(std::move(peek(stack, 0, 4)), size, stride, storage_offset);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto storage_offset = int64_t(node->i(Symbol::attr("storage_offset")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("as_strided");
    
        auto result = at::as_strided(std::move(peek(stack, 0, 1)), size, stride, storage_offset);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::asin(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("asin");
    
        auto result = at::asin(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::atan(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("atan");
    
        auto result = at::atan(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::atan2(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("atan2");
    
        auto result = at::atan2(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::avg_pool1d(Tensor self, int[1] kernel_size, int[1] stride=[], int[1] padding=0, bool ceil_mode=False, bool count_include_pad=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool1d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 4, 6)));
        auto count_include_pad = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::avg_pool1d(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      auto count_include_pad = bool(node->i(Symbol::attr("count_include_pad")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool1d");
    
        auto result = at::avg_pool1d(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::avg_pool2d(Tensor self, int[2] kernel_size, int[2] stride=[], int[2] padding=0, bool ceil_mode=False, bool count_include_pad=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool2d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 4, 6)));
        auto count_include_pad = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::avg_pool2d(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      auto count_include_pad = bool(node->i(Symbol::attr("count_include_pad")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool2d");
    
        auto result = at::avg_pool2d(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::avg_pool2d_backward(Tensor grad_output, Tensor self, int[2] kernel_size, int[2] stride, int[2] padding, bool ceil_mode, bool count_include_pad) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool2d_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 5, 7)));
        auto count_include_pad = tensor_as<bool>(std::move(peek(stack, 6, 7)));
        auto result = at::avg_pool2d_backward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      auto count_include_pad = bool(node->i(Symbol::attr("count_include_pad")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool2d_backward");
    
        auto result = at::avg_pool2d_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::avg_pool2d_forward(Tensor self, int[2] kernel_size, int[2] stride, int[2] padding, bool ceil_mode, bool count_include_pad) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool2d_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 4, 6)));
        auto count_include_pad = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::avg_pool2d_forward(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      auto count_include_pad = bool(node->i(Symbol::attr("count_include_pad")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool2d_forward");
    
        auto result = at::avg_pool2d_forward(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::avg_pool3d(Tensor self, int[3] kernel_size, int[3] stride=[], int[3] padding=0, bool ceil_mode=False, bool count_include_pad=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool3d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 4, 6)));
        auto count_include_pad = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::avg_pool3d(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      auto count_include_pad = bool(node->i(Symbol::attr("count_include_pad")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool3d");
    
        auto result = at::avg_pool3d(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::avg_pool3d_backward(Tensor grad_output, Tensor self, int[3] kernel_size, int[3] stride, int[3] padding, bool ceil_mode, bool count_include_pad) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool3d_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 5, 7)));
        auto count_include_pad = tensor_as<bool>(std::move(peek(stack, 6, 7)));
        auto result = at::avg_pool3d_backward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      auto count_include_pad = bool(node->i(Symbol::attr("count_include_pad")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool3d_backward");
    
        auto result = at::avg_pool3d_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::avg_pool3d_forward(Tensor self, int[3] kernel_size, int[3] stride, int[3] padding, bool ceil_mode, bool count_include_pad) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool3d_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 4, 6)));
        auto count_include_pad = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::avg_pool3d_forward(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      auto count_include_pad = bool(node->i(Symbol::attr("count_include_pad")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("avg_pool3d_forward");
    
        auto result = at::avg_pool3d_forward(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, ceil_mode, count_include_pad);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::baddbmm(Tensor self, Tensor batch1, Tensor batch2, *, Scalar beta=1, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("baddbmm");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 4, 5)));
        auto result = at::baddbmm(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), beta, alpha);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("baddbmm");
    
        auto result = at::baddbmm(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), beta, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::bartlett_window(int window_length, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bartlett_window");
        auto window_length = tensor_as<int64_t>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::bartlett_window(window_length, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto window_length = int64_t(node->i(Symbol::attr("window_length")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bartlett_window");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::bartlett_window(window_length, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::bartlett_window(int window_length, bool periodic, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bartlett_window");
        auto window_length = tensor_as<int64_t>(std::move(peek(stack, 0, 5)));
        auto periodic = tensor_as<bool>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::bartlett_window(window_length, periodic, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto window_length = int64_t(node->i(Symbol::attr("window_length")));
      auto periodic = bool(node->i(Symbol::attr("periodic")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bartlett_window");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::bartlett_window(window_length, periodic, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::batch_norm(Tensor input, Tensor? weight, Tensor? bias, Tensor? running_mean, Tensor? running_var, bool training, float momentum, float eps, bool cudnn_enabled) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("batch_norm");
        auto training = tensor_as<bool>(std::move(peek(stack, 5, 9)));
        auto momentum = tensor_as<double>(std::move(peek(stack, 6, 9)));
        auto eps = tensor_as<double>(std::move(peek(stack, 7, 9)));
        auto cudnn_enabled = tensor_as<bool>(std::move(peek(stack, 8, 9)));
        auto result = at::batch_norm(std::move(peek(stack, 0, 9)), std::move(peek(stack, 1, 9)), std::move(peek(stack, 2, 9)), std::move(peek(stack, 3, 9)), std::move(peek(stack, 4, 9)), training, momentum, eps, cudnn_enabled);
        drop(stack, 9);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto training = bool(node->i(Symbol::attr("training")));
      auto momentum = double(node->f(Symbol::attr("momentum")));
      auto eps = double(node->f(Symbol::attr("eps")));
      auto cudnn_enabled = bool(node->i(Symbol::attr("cudnn_enabled")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("batch_norm");
    
        auto result = at::batch_norm(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), training, momentum, eps, cudnn_enabled);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::bernoulli(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bernoulli");
    
        auto result = at::bernoulli(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::bernoulli(Tensor self, Tensor p, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bernoulli");
    
        auto result = at::bernoulli(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), nullptr);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::bernoulli(Tensor self, float p, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bernoulli");
        auto p = tensor_as<double>(std::move(peek(stack, 1, 2)));
        auto result = at::bernoulli(std::move(peek(stack, 0, 2)), p, nullptr);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = double(node->f(Symbol::attr("p")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bernoulli");
    
        auto result = at::bernoulli(std::move(peek(stack, 0, 1)), p, nullptr);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::bilinear(Tensor input1, Tensor input2, Tensor weight, Tensor? bias) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bilinear");
    
        auto result = at::bilinear(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), std::move(peek(stack, 3, 4)));
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::binary_cross_entropy(Tensor self, Tensor target, Tensor weight=None, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("binary_cross_entropy");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::binary_cross_entropy(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), reduction);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("binary_cross_entropy");
    
        auto result = at::binary_cross_entropy(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::binary_cross_entropy_backward(Tensor grad_output, Tensor self, Tensor target, Tensor? weight, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("binary_cross_entropy_backward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::binary_cross_entropy_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), std::move(peek(stack, 3, 5)), reduction);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("binary_cross_entropy_backward");
    
        auto result = at::binary_cross_entropy_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), std::move(peek(stack, 3, 4)), reduction);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::binary_cross_entropy_forward(Tensor self, Tensor target, Tensor? weight, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("binary_cross_entropy_forward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::binary_cross_entropy_forward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), reduction);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("binary_cross_entropy_forward");
    
        auto result = at::binary_cross_entropy_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::bincount(Tensor self, Tensor? weights=None, int minlength=0) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bincount");
        auto minlength = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::bincount(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), minlength);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto minlength = int64_t(node->i(Symbol::attr("minlength")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bincount");
    
        auto result = at::bincount(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), minlength);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::blackman_window(int window_length, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("blackman_window");
        auto window_length = tensor_as<int64_t>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::blackman_window(window_length, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto window_length = int64_t(node->i(Symbol::attr("window_length")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("blackman_window");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::blackman_window(window_length, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::blackman_window(int window_length, bool periodic, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("blackman_window");
        auto window_length = tensor_as<int64_t>(std::move(peek(stack, 0, 5)));
        auto periodic = tensor_as<bool>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::blackman_window(window_length, periodic, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto window_length = int64_t(node->i(Symbol::attr("window_length")));
      auto periodic = bool(node->i(Symbol::attr("periodic")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("blackman_window");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::blackman_window(window_length, periodic, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::bmm(Tensor self, Tensor mat2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("bmm");
    
        auto result = at::bmm(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::btrifact(Tensor self, *, bool pivot=True) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("btrifact");
        auto pivot = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::btrifact(std::move(peek(stack, 0, 2)), pivot);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto pivot = bool(node->i(Symbol::attr("pivot")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("btrifact");
    
        auto result = at::btrifact(std::move(peek(stack, 0, 1)), pivot);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::btrifact_with_info(Tensor self, *, bool pivot=True) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("btrifact_with_info");
        auto pivot = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::btrifact_with_info(std::move(peek(stack, 0, 2)), pivot);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto pivot = bool(node->i(Symbol::attr("pivot")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("btrifact_with_info");
    
        auto result = at::btrifact_with_info(std::move(peek(stack, 0, 1)), pivot);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::btrisolve(Tensor self, Tensor LU_data, Tensor LU_pivots) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("btrisolve");
    
        auto result = at::btrisolve(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cat(Tensor[] tensors, int dim=0) -> Tensor",
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cat");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 0, 1)));
        auto result = at::cat(peekSlice(stack, 0, varargs_length - 1, varargs_length), dim);
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cat");
    
        auto result = at::cat(peekSlice(stack, 0, varargs_length - 0, varargs_length), dim);
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ceil(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ceil");
    
        auto result = at::ceil(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::chunk(Tensor self, int chunks, int dim=0) -> Tensor[]",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("chunk");
        auto chunks = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::chunk(std::move(peek(stack, 0, 3)), chunks, dim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto chunks = int64_t(node->i(Symbol::attr("chunks")));
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("chunk");
    
        auto result = at::chunk(std::move(peek(stack, 0, 1)), chunks, dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::clamp(Tensor self, Scalar min, Scalar max) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("clamp");
        auto min = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto max = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::clamp(std::move(peek(stack, 0, 3)), min, max);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto min = Scalar(node->t(Symbol::attr("min")));
      auto max = Scalar(node->t(Symbol::attr("max")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("clamp");
    
        auto result = at::clamp(std::move(peek(stack, 0, 1)), min, max);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::clamp_max(Tensor self, Scalar max) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("clamp_max");
        auto max = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::clamp_max(std::move(peek(stack, 0, 2)), max);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto max = Scalar(node->t(Symbol::attr("max")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("clamp_max");
    
        auto result = at::clamp_max(std::move(peek(stack, 0, 1)), max);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::clamp_min(Tensor self, Scalar min) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("clamp_min");
        auto min = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::clamp_min(std::move(peek(stack, 0, 2)), min);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto min = Scalar(node->t(Symbol::attr("min")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("clamp_min");
    
        auto result = at::clamp_min(std::move(peek(stack, 0, 1)), min);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::clone(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("clone");
    
        auto result = at::clone(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::coalesce(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("coalesce");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).coalesce();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::contiguous(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("contiguous");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).contiguous();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::conv1d(Tensor input, Tensor weight, Tensor bias=None, int[1] stride=1, int[1] padding=0, int[1] dilation=1, int groups=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv1d");
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 7)));
        auto result = at::conv1d(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), stride, padding, dilation, groups);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv1d");
    
        auto result = at::conv1d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), stride, padding, dilation, groups);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::conv2d(Tensor input, Tensor weight, Tensor bias=None, int[2] stride=1, int[2] padding=0, int[2] dilation=1, int groups=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv2d");
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 7)));
        auto result = at::conv2d(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), stride, padding, dilation, groups);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv2d");
    
        auto result = at::conv2d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), stride, padding, dilation, groups);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::conv3d(Tensor input, Tensor weight, Tensor bias=None, int[3] stride=1, int[3] padding=0, int[3] dilation=1, int groups=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv3d");
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 7)));
        auto result = at::conv3d(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), stride, padding, dilation, groups);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv3d");
    
        auto result = at::conv3d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), stride, padding, dilation, groups);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::conv_tbc(Tensor self, Tensor weight, Tensor bias, int pad) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv_tbc");
        auto pad = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::conv_tbc(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), pad);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto pad = int64_t(node->i(Symbol::attr("pad")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv_tbc");
    
        auto result = at::conv_tbc(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), pad);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::conv_tbc_backward(Tensor self, Tensor input, Tensor weight, Tensor bias, int pad) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv_tbc_backward");
        auto pad = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::conv_tbc_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), std::move(peek(stack, 3, 5)), pad);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto pad = int64_t(node->i(Symbol::attr("pad")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv_tbc_backward");
    
        auto result = at::conv_tbc_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), std::move(peek(stack, 3, 4)), pad);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::conv_transpose1d(Tensor input, Tensor weight, Tensor bias=None, int[1] stride=1, int[1] padding=0, int[1] output_padding=0, int groups=1, int[1] dilation=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv_transpose1d");
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 7, 8)));
        auto result = at::conv_transpose1d(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), stride, padding, output_padding, groups, dilation);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv_transpose1d");
    
        auto result = at::conv_transpose1d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), stride, padding, output_padding, groups, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::conv_transpose2d(Tensor input, Tensor weight, Tensor bias=None, int[2] stride=1, int[2] padding=0, int[2] output_padding=0, int groups=1, int[2] dilation=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv_transpose2d");
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 7, 8)));
        auto result = at::conv_transpose2d(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), stride, padding, output_padding, groups, dilation);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv_transpose2d");
    
        auto result = at::conv_transpose2d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), stride, padding, output_padding, groups, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::conv_transpose3d(Tensor input, Tensor weight, Tensor bias=None, int[3] stride=1, int[3] padding=0, int[3] output_padding=0, int groups=1, int[3] dilation=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv_transpose3d");
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 7, 8)));
        auto result = at::conv_transpose3d(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), stride, padding, output_padding, groups, dilation);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("conv_transpose3d");
    
        auto result = at::conv_transpose3d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), stride, padding, output_padding, groups, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::convolution(Tensor input, Tensor weight, Tensor? bias, int[] stride, int[] padding, int[] dilation, bool transposed, int[] output_padding, int groups) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("convolution");
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 9)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 9)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 9)));
        auto transposed = tensor_as<bool>(std::move(peek(stack, 6, 9)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 7, 9)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 8, 9)));
        auto result = at::convolution(std::move(peek(stack, 0, 9)), std::move(peek(stack, 1, 9)), std::move(peek(stack, 2, 9)), stride, padding, dilation, transposed, output_padding, groups);
        drop(stack, 9);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto transposed = bool(node->i(Symbol::attr("transposed")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("convolution");
    
        auto result = at::convolution(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), stride, padding, dilation, transposed, output_padding, groups);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cos(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cos");
    
        auto result = at::cos(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cosh(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cosh");
    
        auto result = at::cosh(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cosine_embedding_loss(Tensor input1, Tensor input2, Tensor target, float margin=0.0, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cosine_embedding_loss");
        auto margin = tensor_as<double>(std::move(peek(stack, 3, 5)));
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::cosine_embedding_loss(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), margin, reduction);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto margin = double(node->f(Symbol::attr("margin")));
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cosine_embedding_loss");
    
        auto result = at::cosine_embedding_loss(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), margin, reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cross(Tensor self, Tensor other, int dim=-1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cross");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::cross(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), dim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cross");
    
        auto result = at::cross(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_affine_grid_generator(Tensor theta, int N, int C, int H, int W) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_affine_grid_generator");
        auto N = tensor_as<int64_t>(std::move(peek(stack, 1, 5)));
        auto C = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto H = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto W = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::cudnn_affine_grid_generator(std::move(peek(stack, 0, 5)), N, C, H, W);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto N = int64_t(node->i(Symbol::attr("N")));
      auto C = int64_t(node->i(Symbol::attr("C")));
      auto H = int64_t(node->i(Symbol::attr("H")));
      auto W = int64_t(node->i(Symbol::attr("W")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_affine_grid_generator");
    
        auto result = at::cudnn_affine_grid_generator(std::move(peek(stack, 0, 1)), N, C, H, W);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_affine_grid_generator_backward(Tensor grad, int N, int C, int H, int W) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_affine_grid_generator_backward");
        auto N = tensor_as<int64_t>(std::move(peek(stack, 1, 5)));
        auto C = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto H = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto W = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::cudnn_affine_grid_generator_backward(std::move(peek(stack, 0, 5)), N, C, H, W);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto N = int64_t(node->i(Symbol::attr("N")));
      auto C = int64_t(node->i(Symbol::attr("C")));
      auto H = int64_t(node->i(Symbol::attr("H")));
      auto W = int64_t(node->i(Symbol::attr("W")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_affine_grid_generator_backward");
    
        auto result = at::cudnn_affine_grid_generator_backward(std::move(peek(stack, 0, 1)), N, C, H, W);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_batch_norm(Tensor input, Tensor weight, Tensor? bias, Tensor? running_mean, Tensor? running_var, bool training, float exponential_average_factor, float epsilon) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_batch_norm");
        auto training = tensor_as<bool>(std::move(peek(stack, 5, 8)));
        auto exponential_average_factor = tensor_as<double>(std::move(peek(stack, 6, 8)));
        auto epsilon = tensor_as<double>(std::move(peek(stack, 7, 8)));
        auto result = at::cudnn_batch_norm(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), std::move(peek(stack, 3, 8)), std::move(peek(stack, 4, 8)), training, exponential_average_factor, epsilon);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto training = bool(node->i(Symbol::attr("training")));
      auto exponential_average_factor = double(node->f(Symbol::attr("exponential_average_factor")));
      auto epsilon = double(node->f(Symbol::attr("epsilon")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_batch_norm");
    
        auto result = at::cudnn_batch_norm(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), training, exponential_average_factor, epsilon);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_batch_norm_backward(Tensor input, Tensor grad_output, Tensor weight, Tensor? running_mean, Tensor? running_var, Tensor? save_mean, Tensor? save_var, float epsilon) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_batch_norm_backward");
        auto epsilon = tensor_as<double>(std::move(peek(stack, 7, 8)));
        auto result = at::cudnn_batch_norm_backward(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), std::move(peek(stack, 3, 8)), std::move(peek(stack, 4, 8)), std::move(peek(stack, 5, 8)), std::move(peek(stack, 6, 8)), epsilon);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto epsilon = double(node->f(Symbol::attr("epsilon")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_batch_norm_backward");
    
        auto result = at::cudnn_batch_norm_backward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), std::move(peek(stack, 3, 7)), std::move(peek(stack, 4, 7)), std::move(peek(stack, 5, 7)), std::move(peek(stack, 6, 7)), epsilon);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_convolution(Tensor self, Tensor weight, Tensor? bias, int[] padding, int[] stride, int[] dilation, int groups, bool benchmark, bool deterministic) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 9)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 9)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 9)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 9)));
        auto benchmark = tensor_as<bool>(std::move(peek(stack, 7, 9)));
        auto deterministic = tensor_as<bool>(std::move(peek(stack, 8, 9)));
        auto result = at::cudnn_convolution(std::move(peek(stack, 0, 9)), std::move(peek(stack, 1, 9)), std::move(peek(stack, 2, 9)), padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 9);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto benchmark = bool(node->i(Symbol::attr("benchmark")));
      auto deterministic = bool(node->i(Symbol::attr("deterministic")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution");
    
        auto result = at::cudnn_convolution(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_convolution_backward(Tensor self, Tensor grad_output, Tensor weight, int[] padding, int[] stride, int[] dilation, int groups, bool benchmark, bool deterministic, bool[3] output_mask) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_backward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 10)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 10)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 10)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 10)));
        auto benchmark = tensor_as<bool>(std::move(peek(stack, 7, 10)));
        auto deterministic = tensor_as<bool>(std::move(peek(stack, 8, 10)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 9, 10)));
        auto result = at::cudnn_convolution_backward(std::move(peek(stack, 0, 10)), std::move(peek(stack, 1, 10)), std::move(peek(stack, 2, 10)), padding, stride, dilation, groups, benchmark, deterministic, output_mask);
        drop(stack, 10);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto benchmark = bool(node->i(Symbol::attr("benchmark")));
      auto deterministic = bool(node->i(Symbol::attr("deterministic")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_backward");
    
        auto result = at::cudnn_convolution_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), padding, stride, dilation, groups, benchmark, deterministic, output_mask);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_convolution_backward_bias(Tensor grad_output) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_backward_bias");
    
        auto result = at::cudnn_convolution_backward_bias(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_convolution_backward_input(int[] self_size, Tensor grad_output, Tensor weight, int[] padding, int[] stride, int[] dilation, int groups, bool benchmark, bool deterministic) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_backward_input");
        auto self_size = tensor_as<IntList>(std::move(peek(stack, 0, 9)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 9)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 9)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 9)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 9)));
        auto benchmark = tensor_as<bool>(std::move(peek(stack, 7, 9)));
        auto deterministic = tensor_as<bool>(std::move(peek(stack, 8, 9)));
        auto result = at::cudnn_convolution_backward_input(self_size, std::move(peek(stack, 1, 9)), std::move(peek(stack, 2, 9)), padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 9);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto self_size = std::vector<int64_t>(node->is(Symbol::attr("self_size")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto benchmark = bool(node->i(Symbol::attr("benchmark")));
      auto deterministic = bool(node->i(Symbol::attr("deterministic")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_backward_input");
    
        auto result = at::cudnn_convolution_backward_input(self_size, std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_convolution_backward_weight(int[] weight_size, Tensor grad_output, Tensor self, int[] padding, int[] stride, int[] dilation, int groups, bool benchmark, bool deterministic) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_backward_weight");
        auto weight_size = tensor_as<IntList>(std::move(peek(stack, 0, 9)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 9)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 9)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 9)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 9)));
        auto benchmark = tensor_as<bool>(std::move(peek(stack, 7, 9)));
        auto deterministic = tensor_as<bool>(std::move(peek(stack, 8, 9)));
        auto result = at::cudnn_convolution_backward_weight(weight_size, std::move(peek(stack, 1, 9)), std::move(peek(stack, 2, 9)), padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 9);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto weight_size = std::vector<int64_t>(node->is(Symbol::attr("weight_size")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto benchmark = bool(node->i(Symbol::attr("benchmark")));
      auto deterministic = bool(node->i(Symbol::attr("deterministic")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_backward_weight");
    
        auto result = at::cudnn_convolution_backward_weight(weight_size, std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_convolution_transpose(Tensor self, Tensor weight, Tensor? bias, int[] padding, int[] output_padding, int[] stride, int[] dilation, int groups, bool benchmark, bool deterministic) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_transpose");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 10)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 4, 10)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 5, 10)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 10)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 7, 10)));
        auto benchmark = tensor_as<bool>(std::move(peek(stack, 8, 10)));
        auto deterministic = tensor_as<bool>(std::move(peek(stack, 9, 10)));
        auto result = at::cudnn_convolution_transpose(std::move(peek(stack, 0, 10)), std::move(peek(stack, 1, 10)), std::move(peek(stack, 2, 10)), padding, output_padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 10);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto benchmark = bool(node->i(Symbol::attr("benchmark")));
      auto deterministic = bool(node->i(Symbol::attr("deterministic")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_transpose");
    
        auto result = at::cudnn_convolution_transpose(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), padding, output_padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_convolution_transpose_backward(Tensor self, Tensor grad_output, Tensor weight, int[] padding, int[] output_padding, int[] stride, int[] dilation, int groups, bool benchmark, bool deterministic, bool[3] output_mask) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_transpose_backward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 11)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 4, 11)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 5, 11)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 11)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 7, 11)));
        auto benchmark = tensor_as<bool>(std::move(peek(stack, 8, 11)));
        auto deterministic = tensor_as<bool>(std::move(peek(stack, 9, 11)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 10, 11)));
        auto result = at::cudnn_convolution_transpose_backward(std::move(peek(stack, 0, 11)), std::move(peek(stack, 1, 11)), std::move(peek(stack, 2, 11)), padding, output_padding, stride, dilation, groups, benchmark, deterministic, output_mask);
        drop(stack, 11);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto benchmark = bool(node->i(Symbol::attr("benchmark")));
      auto deterministic = bool(node->i(Symbol::attr("deterministic")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_transpose_backward");
    
        auto result = at::cudnn_convolution_transpose_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), padding, output_padding, stride, dilation, groups, benchmark, deterministic, output_mask);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_convolution_transpose_backward_bias(Tensor grad_output) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_transpose_backward_bias");
    
        auto result = at::cudnn_convolution_transpose_backward_bias(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_convolution_transpose_backward_input(Tensor grad_output, Tensor weight, int[] padding, int[] stride, int[] dilation, int groups, bool benchmark, bool deterministic) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_transpose_backward_input");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 2, 8)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 5, 8)));
        auto benchmark = tensor_as<bool>(std::move(peek(stack, 6, 8)));
        auto deterministic = tensor_as<bool>(std::move(peek(stack, 7, 8)));
        auto result = at::cudnn_convolution_transpose_backward_input(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto benchmark = bool(node->i(Symbol::attr("benchmark")));
      auto deterministic = bool(node->i(Symbol::attr("deterministic")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_transpose_backward_input");
    
        auto result = at::cudnn_convolution_transpose_backward_input(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_convolution_transpose_backward_weight(int[] weight_size, Tensor grad_output, Tensor self, int[] padding, int[] stride, int[] dilation, int groups, bool benchmark, bool deterministic) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_transpose_backward_weight");
        auto weight_size = tensor_as<IntList>(std::move(peek(stack, 0, 9)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 9)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 9)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 9)));
        auto groups = tensor_as<int64_t>(std::move(peek(stack, 6, 9)));
        auto benchmark = tensor_as<bool>(std::move(peek(stack, 7, 9)));
        auto deterministic = tensor_as<bool>(std::move(peek(stack, 8, 9)));
        auto result = at::cudnn_convolution_transpose_backward_weight(weight_size, std::move(peek(stack, 1, 9)), std::move(peek(stack, 2, 9)), padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 9);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto weight_size = std::vector<int64_t>(node->is(Symbol::attr("weight_size")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto groups = int64_t(node->i(Symbol::attr("groups")));
      auto benchmark = bool(node->i(Symbol::attr("benchmark")));
      auto deterministic = bool(node->i(Symbol::attr("deterministic")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_convolution_transpose_backward_weight");
    
        auto result = at::cudnn_convolution_transpose_backward_weight(weight_size, std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding, stride, dilation, groups, benchmark, deterministic);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_grid_sampler(Tensor self, Tensor grid) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_grid_sampler");
    
        auto result = at::cudnn_grid_sampler(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_grid_sampler_backward(Tensor self, Tensor grid, Tensor grad_output) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_grid_sampler_backward");
    
        auto result = at::cudnn_grid_sampler_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cudnn_is_acceptable(Tensor self) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cudnn_is_acceptable");
    
        auto result = at::cudnn_is_acceptable(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cumprod(Tensor self, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cumprod");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::cumprod(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cumprod");
    
        auto result = at::cumprod(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::cumsum(Tensor self, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cumsum");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::cumsum(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("cumsum");
    
        auto result = at::cumsum(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::det(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("det");
    
        auto result = at::det(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::diag(Tensor self, int diagonal=0) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("diag");
        auto diagonal = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::diag(std::move(peek(stack, 0, 2)), diagonal);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto diagonal = int64_t(node->i(Symbol::attr("diagonal")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("diag");
    
        auto result = at::diag(std::move(peek(stack, 0, 1)), diagonal);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::diagflat(Tensor self, int offset=0) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("diagflat");
        auto offset = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::diagflat(std::move(peek(stack, 0, 2)), offset);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto offset = int64_t(node->i(Symbol::attr("offset")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("diagflat");
    
        auto result = at::diagflat(std::move(peek(stack, 0, 1)), offset);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::diagonal(Tensor self, int offset=0, int dim1=0, int dim2=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("diagonal");
        auto offset = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto dim1 = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto dim2 = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::diagonal(std::move(peek(stack, 0, 4)), offset, dim1, dim2);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto offset = int64_t(node->i(Symbol::attr("offset")));
      auto dim1 = int64_t(node->i(Symbol::attr("dim1")));
      auto dim2 = int64_t(node->i(Symbol::attr("dim2")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("diagonal");
    
        auto result = at::diagonal(std::move(peek(stack, 0, 1)), offset, dim1, dim2);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::digamma(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("digamma");
    
        auto result = at::digamma(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::dim(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("dim");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).dim();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::dist(Tensor self, Tensor other, Scalar p=2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("dist");
        auto p = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::dist(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), p);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = Scalar(node->t(Symbol::attr("p")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("dist");
    
        auto result = at::dist(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), p);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::div(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("div");
    
        auto result = at::div(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::div(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("div");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::div(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("div");
    
        auto result = at::div(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::dot(Tensor self, Tensor tensor) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("dot");
    
        auto result = at::dot(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::eig(Tensor self, bool eigenvectors=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("eig");
        auto eigenvectors = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::eig(std::move(peek(stack, 0, 2)), eigenvectors);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto eigenvectors = bool(node->i(Symbol::attr("eigenvectors")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("eig");
    
        auto result = at::eig(std::move(peek(stack, 0, 1)), eigenvectors);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::elu(Tensor self, Scalar alpha=1, Scalar scale=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("elu");
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto scale = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::elu(std::move(peek(stack, 0, 3)), alpha, scale);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      auto scale = Scalar(node->t(Symbol::attr("scale")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("elu");
    
        auto result = at::elu(std::move(peek(stack, 0, 1)), alpha, scale);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::elu_backward(Tensor grad_output, Scalar alpha, Scalar scale, Tensor output) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("elu_backward");
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 1, 4)));
        auto scale = tensor_as<Scalar>(std::move(peek(stack, 2, 4)));
        auto result = at::elu_backward(std::move(peek(stack, 0, 4)), alpha, scale, std::move(peek(stack, 3, 4)));
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      auto scale = Scalar(node->t(Symbol::attr("scale")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("elu_backward");
    
        auto result = at::elu_backward(std::move(peek(stack, 0, 2)), alpha, scale, std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::elu_forward(Tensor self, Scalar alpha, Scalar scale) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("elu_forward");
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto scale = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::elu_forward(std::move(peek(stack, 0, 3)), alpha, scale);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      auto scale = Scalar(node->t(Symbol::attr("scale")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("elu_forward");
    
        auto result = at::elu_forward(std::move(peek(stack, 0, 1)), alpha, scale);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::embedding(Tensor weight, Tensor indices, int padding_idx=-1, bool scale_grad_by_freq=False, bool sparse=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("embedding");
        auto padding_idx = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto scale_grad_by_freq = tensor_as<bool>(std::move(peek(stack, 3, 5)));
        auto sparse = tensor_as<bool>(std::move(peek(stack, 4, 5)));
        auto result = at::embedding(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), padding_idx, scale_grad_by_freq, sparse);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding_idx = int64_t(node->i(Symbol::attr("padding_idx")));
      auto scale_grad_by_freq = bool(node->i(Symbol::attr("scale_grad_by_freq")));
      auto sparse = bool(node->i(Symbol::attr("sparse")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("embedding");
    
        auto result = at::embedding(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding_idx, scale_grad_by_freq, sparse);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::embedding_backward(Tensor grad, Tensor indices, int num_weights, int padding_idx, bool scale_grad_by_freq, bool sparse) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("embedding_backward");
        auto num_weights = tensor_as<int64_t>(std::move(peek(stack, 2, 6)));
        auto padding_idx = tensor_as<int64_t>(std::move(peek(stack, 3, 6)));
        auto scale_grad_by_freq = tensor_as<bool>(std::move(peek(stack, 4, 6)));
        auto sparse = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::embedding_backward(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), num_weights, padding_idx, scale_grad_by_freq, sparse);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto num_weights = int64_t(node->i(Symbol::attr("num_weights")));
      auto padding_idx = int64_t(node->i(Symbol::attr("padding_idx")));
      auto scale_grad_by_freq = bool(node->i(Symbol::attr("scale_grad_by_freq")));
      auto sparse = bool(node->i(Symbol::attr("sparse")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("embedding_backward");
    
        auto result = at::embedding_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), num_weights, padding_idx, scale_grad_by_freq, sparse);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::embedding_bag(Tensor weight, Tensor indices, Tensor offsets, bool scale_grad_by_freq=False, int mode=0, bool sparse=False) -> (Tensor, Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("embedding_bag");
        auto scale_grad_by_freq = tensor_as<bool>(std::move(peek(stack, 3, 6)));
        auto mode = tensor_as<int64_t>(std::move(peek(stack, 4, 6)));
        auto sparse = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::embedding_bag(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), std::move(peek(stack, 2, 6)), scale_grad_by_freq, mode, sparse);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto scale_grad_by_freq = bool(node->i(Symbol::attr("scale_grad_by_freq")));
      auto mode = int64_t(node->i(Symbol::attr("mode")));
      auto sparse = bool(node->i(Symbol::attr("sparse")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("embedding_bag");
    
        auto result = at::embedding_bag(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), scale_grad_by_freq, mode, sparse);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::embedding_dense_backward(Tensor grad, Tensor indices, int num_weights, int padding_idx, bool scale_grad_by_freq) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("embedding_dense_backward");
        auto num_weights = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto padding_idx = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto scale_grad_by_freq = tensor_as<bool>(std::move(peek(stack, 4, 5)));
        auto result = at::embedding_dense_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), num_weights, padding_idx, scale_grad_by_freq);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto num_weights = int64_t(node->i(Symbol::attr("num_weights")));
      auto padding_idx = int64_t(node->i(Symbol::attr("padding_idx")));
      auto scale_grad_by_freq = bool(node->i(Symbol::attr("scale_grad_by_freq")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("embedding_dense_backward");
    
        auto result = at::embedding_dense_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), num_weights, padding_idx, scale_grad_by_freq);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::embedding_sparse_backward(Tensor grad, Tensor indices, int num_weights, int padding_idx, bool scale_grad_by_freq) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("embedding_sparse_backward");
        auto num_weights = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto padding_idx = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto scale_grad_by_freq = tensor_as<bool>(std::move(peek(stack, 4, 5)));
        auto result = at::embedding_sparse_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), num_weights, padding_idx, scale_grad_by_freq);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto num_weights = int64_t(node->i(Symbol::attr("num_weights")));
      auto padding_idx = int64_t(node->i(Symbol::attr("padding_idx")));
      auto scale_grad_by_freq = bool(node->i(Symbol::attr("scale_grad_by_freq")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("embedding_sparse_backward");
    
        auto result = at::embedding_sparse_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), num_weights, padding_idx, scale_grad_by_freq);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::empty(int[] size, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("empty");
        auto size = tensor_as<IntList>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::empty(size, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("empty");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::empty(size, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::empty_like(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("empty_like");
    
        auto result = at::empty_like(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::empty_like(Tensor self, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("empty_like");
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::empty_like(std::move(peek(stack, 0, 4)), options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("empty_like");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::empty_like(std::move(peek(stack, 0, 1)), options);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::eq(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("eq");
    
        auto result = at::eq(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::eq(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("eq");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::eq(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("eq");
    
        auto result = at::eq(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::equal(Tensor self, Tensor other) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("equal");
    
        auto result = at::equal(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::erf(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("erf");
    
        auto result = at::erf(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::erfinv(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("erfinv");
    
        auto result = at::erfinv(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::exp(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("exp");
    
        auto result = at::exp(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::expand(Tensor self, int[] size, *, bool implicit=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("expand");
        auto size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto implicit = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        DeviceGuard device_guard(deviceForInputs(stack, 3));
        auto result = (std::move(peek(stack, 0, 3))).expand(size, implicit);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto implicit = bool(node->i(Symbol::attr("implicit")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("expand");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).expand(size, implicit);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::expand_as(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("expand_as");
    
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).expand_as(std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::expm1(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("expm1");
    
        auto result = at::expm1(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::eye(int n, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("eye");
        auto n = tensor_as<int64_t>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::eye(n, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto n = int64_t(node->i(Symbol::attr("n")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("eye");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::eye(n, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::eye(int n, int m, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("eye");
        auto n = tensor_as<int64_t>(std::move(peek(stack, 0, 5)));
        auto m = tensor_as<int64_t>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::eye(n, m, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto n = int64_t(node->i(Symbol::attr("n")));
      auto m = int64_t(node->i(Symbol::attr("m")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("eye");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::eye(n, m, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::fft(Tensor self, int signal_ndim, bool normalized=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fft");
        auto signal_ndim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto normalized = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::fft(std::move(peek(stack, 0, 3)), signal_ndim, normalized);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto signal_ndim = int64_t(node->i(Symbol::attr("signal_ndim")));
      auto normalized = bool(node->i(Symbol::attr("normalized")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fft");
    
        auto result = at::fft(std::move(peek(stack, 0, 1)), signal_ndim, normalized);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::flatten(Tensor self, int start_dim=0, int end_dim=-1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("flatten");
        auto start_dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto end_dim = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::flatten(std::move(peek(stack, 0, 3)), start_dim, end_dim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto start_dim = int64_t(node->i(Symbol::attr("start_dim")));
      auto end_dim = int64_t(node->i(Symbol::attr("end_dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("flatten");
    
        auto result = at::flatten(std::move(peek(stack, 0, 1)), start_dim, end_dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::flip(Tensor self, int[] dims) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("flip");
        auto dims = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::flip(std::move(peek(stack, 0, 2)), dims);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dims = std::vector<int64_t>(node->is(Symbol::attr("dims")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("flip");
    
        auto result = at::flip(std::move(peek(stack, 0, 1)), dims);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::floor(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("floor");
    
        auto result = at::floor(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::fmod(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fmod");
    
        auto result = at::fmod(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::fmod(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fmod");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::fmod(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fmod");
    
        auto result = at::fmod(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::frac(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("frac");
    
        auto result = at::frac(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::fractional_max_pool2d(Tensor self, int[2] kernel_size, int[2] output_size, Tensor random_samples) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fractional_max_pool2d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 4)));
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 2, 4)));
        auto result = at::fractional_max_pool2d(std::move(peek(stack, 0, 4)), kernel_size, output_size, std::move(peek(stack, 3, 4)));
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fractional_max_pool2d");
    
        auto result = at::fractional_max_pool2d(std::move(peek(stack, 0, 2)), kernel_size, output_size, std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::fractional_max_pool2d_backward(Tensor grad_output, Tensor self, int[2] kernel_size, int[2] output_size, Tensor indices) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fractional_max_pool2d_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 5)));
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 3, 5)));
        auto result = at::fractional_max_pool2d_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), kernel_size, output_size, std::move(peek(stack, 4, 5)));
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fractional_max_pool2d_backward");
    
        auto result = at::fractional_max_pool2d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, output_size, std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::fractional_max_pool2d_forward(Tensor self, int[2] kernel_size, int[2] output_size, Tensor random_samples) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fractional_max_pool2d_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 4)));
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 2, 4)));
        auto result = at::fractional_max_pool2d_forward(std::move(peek(stack, 0, 4)), kernel_size, output_size, std::move(peek(stack, 3, 4)));
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("fractional_max_pool2d_forward");
    
        auto result = at::fractional_max_pool2d_forward(std::move(peek(stack, 0, 2)), kernel_size, output_size, std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::full(int[] size, Scalar fill_value, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("full");
        auto size = tensor_as<IntList>(std::move(peek(stack, 0, 5)));
        auto fill_value = tensor_as<Scalar>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::full(size, fill_value, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto fill_value = Scalar(node->t(Symbol::attr("fill_value")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("full");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::full(size, fill_value, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::full_like(Tensor self, Scalar fill_value) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("full_like");
        auto fill_value = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::full_like(std::move(peek(stack, 0, 2)), fill_value);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto fill_value = Scalar(node->t(Symbol::attr("fill_value")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("full_like");
    
        auto result = at::full_like(std::move(peek(stack, 0, 1)), fill_value);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::full_like(Tensor self, Scalar fill_value, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("full_like");
        auto fill_value = tensor_as<Scalar>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::full_like(std::move(peek(stack, 0, 5)), fill_value, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto fill_value = Scalar(node->t(Symbol::attr("fill_value")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("full_like");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::full_like(std::move(peek(stack, 0, 1)), fill_value, options);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::gather(Tensor self, int dim, Tensor index) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("gather");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto result = at::gather(std::move(peek(stack, 0, 3)), dim, std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("gather");
    
        auto result = at::gather(std::move(peek(stack, 0, 2)), dim, std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ge(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ge");
    
        auto result = at::ge(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ge(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ge");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::ge(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ge");
    
        auto result = at::ge(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::gels(Tensor self, Tensor A) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("gels");
    
        auto result = at::gels(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::geqrf(Tensor self) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("geqrf");
    
        auto result = at::geqrf(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ger(Tensor self, Tensor vec2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ger");
    
        auto result = at::ger(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::gesv(Tensor self, Tensor A) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("gesv");
    
        auto result = at::gesv(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::get_device(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("get_device");
    
        auto result = at::get_device(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::glu(Tensor self, int dim=-1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("glu");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::glu(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("glu");
    
        auto result = at::glu(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::glu_backward(Tensor grad_output, Tensor self, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("glu_backward");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::glu_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), dim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("glu_backward");
    
        auto result = at::glu_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::glu_forward(Tensor self, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("glu_forward");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::glu_forward(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("glu_forward");
    
        auto result = at::glu_forward(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::group_norm(Tensor input, int num_groups, Tensor? weight=None, Tensor? bias=None, float eps=1e-05, bool cudnn_enabled=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("group_norm");
        auto num_groups = tensor_as<int64_t>(std::move(peek(stack, 1, 6)));
        auto eps = tensor_as<double>(std::move(peek(stack, 4, 6)));
        auto cudnn_enabled = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::group_norm(std::move(peek(stack, 0, 6)), num_groups, std::move(peek(stack, 2, 6)), std::move(peek(stack, 3, 6)), eps, cudnn_enabled);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto num_groups = int64_t(node->i(Symbol::attr("num_groups")));
      auto eps = double(node->f(Symbol::attr("eps")));
      auto cudnn_enabled = bool(node->i(Symbol::attr("cudnn_enabled")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("group_norm");
    
        auto result = at::group_norm(std::move(peek(stack, 0, 3)), num_groups, std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), eps, cudnn_enabled);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::gt(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("gt");
    
        auto result = at::gt(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::gt(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("gt");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::gt(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("gt");
    
        auto result = at::gt(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hamming_window(int window_length, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hamming_window");
        auto window_length = tensor_as<int64_t>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hamming_window(window_length, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto window_length = int64_t(node->i(Symbol::attr("window_length")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hamming_window");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hamming_window(window_length, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hamming_window(int window_length, bool periodic, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hamming_window");
        auto window_length = tensor_as<int64_t>(std::move(peek(stack, 0, 5)));
        auto periodic = tensor_as<bool>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hamming_window(window_length, periodic, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto window_length = int64_t(node->i(Symbol::attr("window_length")));
      auto periodic = bool(node->i(Symbol::attr("periodic")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hamming_window");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hamming_window(window_length, periodic, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hamming_window(int window_length, bool periodic, float alpha, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hamming_window");
        auto window_length = tensor_as<int64_t>(std::move(peek(stack, 0, 6)));
        auto periodic = tensor_as<bool>(std::move(peek(stack, 1, 6)));
        auto alpha = tensor_as<double>(std::move(peek(stack, 2, 6)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 3, 6)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 4, 6)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hamming_window(window_length, periodic, alpha, options);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto window_length = int64_t(node->i(Symbol::attr("window_length")));
      auto periodic = bool(node->i(Symbol::attr("periodic")));
      auto alpha = double(node->f(Symbol::attr("alpha")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hamming_window");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hamming_window(window_length, periodic, alpha, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hamming_window(int window_length, bool periodic, float alpha, float beta, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hamming_window");
        auto window_length = tensor_as<int64_t>(std::move(peek(stack, 0, 7)));
        auto periodic = tensor_as<bool>(std::move(peek(stack, 1, 7)));
        auto alpha = tensor_as<double>(std::move(peek(stack, 2, 7)));
        auto beta = tensor_as<double>(std::move(peek(stack, 3, 7)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 4, 7)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 5, 7)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 6, 7)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hamming_window(window_length, periodic, alpha, beta, options);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto window_length = int64_t(node->i(Symbol::attr("window_length")));
      auto periodic = bool(node->i(Symbol::attr("periodic")));
      auto alpha = double(node->f(Symbol::attr("alpha")));
      auto beta = double(node->f(Symbol::attr("beta")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hamming_window");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hamming_window(window_length, periodic, alpha, beta, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hann_window(int window_length, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hann_window");
        auto window_length = tensor_as<int64_t>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hann_window(window_length, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto window_length = int64_t(node->i(Symbol::attr("window_length")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hann_window");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hann_window(window_length, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hann_window(int window_length, bool periodic, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hann_window");
        auto window_length = tensor_as<int64_t>(std::move(peek(stack, 0, 5)));
        auto periodic = tensor_as<bool>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hann_window(window_length, periodic, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto window_length = int64_t(node->i(Symbol::attr("window_length")));
      auto periodic = bool(node->i(Symbol::attr("periodic")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hann_window");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::hann_window(window_length, periodic, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hardshrink(Tensor self, Scalar lambd=0.5) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hardshrink");
        auto lambd = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::hardshrink(std::move(peek(stack, 0, 2)), lambd);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto lambd = Scalar(node->t(Symbol::attr("lambd")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hardshrink");
    
        auto result = at::hardshrink(std::move(peek(stack, 0, 1)), lambd);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hardshrink_backward(Tensor grad_out, Tensor self, Scalar lambd) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hardshrink_backward");
        auto lambd = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::hardshrink_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), lambd);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto lambd = Scalar(node->t(Symbol::attr("lambd")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hardshrink_backward");
    
        auto result = at::hardshrink_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), lambd);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hardtanh(Tensor self, Scalar min_val=-1, Scalar max_val=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hardtanh");
        auto min_val = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto max_val = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::hardtanh(std::move(peek(stack, 0, 3)), min_val, max_val);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto min_val = Scalar(node->t(Symbol::attr("min_val")));
      auto max_val = Scalar(node->t(Symbol::attr("max_val")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hardtanh");
    
        auto result = at::hardtanh(std::move(peek(stack, 0, 1)), min_val, max_val);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hardtanh_backward(Tensor grad_output, Tensor self, Scalar min_val, Scalar max_val) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hardtanh_backward");
        auto min_val = tensor_as<Scalar>(std::move(peek(stack, 2, 4)));
        auto max_val = tensor_as<Scalar>(std::move(peek(stack, 3, 4)));
        auto result = at::hardtanh_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), min_val, max_val);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto min_val = Scalar(node->t(Symbol::attr("min_val")));
      auto max_val = Scalar(node->t(Symbol::attr("max_val")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hardtanh_backward");
    
        auto result = at::hardtanh_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), min_val, max_val);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hardtanh_forward(Tensor self, Scalar min_val, Scalar max_val) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hardtanh_forward");
        auto min_val = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto max_val = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::hardtanh_forward(std::move(peek(stack, 0, 3)), min_val, max_val);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto min_val = Scalar(node->t(Symbol::attr("min_val")));
      auto max_val = Scalar(node->t(Symbol::attr("max_val")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hardtanh_forward");
    
        auto result = at::hardtanh_forward(std::move(peek(stack, 0, 1)), min_val, max_val);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hinge_embedding_loss(Tensor self, Tensor target, float margin=1.0, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hinge_embedding_loss");
        auto margin = tensor_as<double>(std::move(peek(stack, 2, 4)));
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::hinge_embedding_loss(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), margin, reduction);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto margin = double(node->f(Symbol::attr("margin")));
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hinge_embedding_loss");
    
        auto result = at::hinge_embedding_loss(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), margin, reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::histc(Tensor self, int bins=100, Scalar min=0, Scalar max=0) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("histc");
        auto bins = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto min = tensor_as<Scalar>(std::move(peek(stack, 2, 4)));
        auto max = tensor_as<Scalar>(std::move(peek(stack, 3, 4)));
        auto result = at::histc(std::move(peek(stack, 0, 4)), bins, min, max);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto bins = int64_t(node->i(Symbol::attr("bins")));
      auto min = Scalar(node->t(Symbol::attr("min")));
      auto max = Scalar(node->t(Symbol::attr("max")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("histc");
    
        auto result = at::histc(std::move(peek(stack, 0, 1)), bins, min, max);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::hspmm(Tensor mat1, Tensor mat2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("hspmm");
    
        auto result = at::hspmm(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ifft(Tensor self, int signal_ndim, bool normalized=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ifft");
        auto signal_ndim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto normalized = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::ifft(std::move(peek(stack, 0, 3)), signal_ndim, normalized);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto signal_ndim = int64_t(node->i(Symbol::attr("signal_ndim")));
      auto normalized = bool(node->i(Symbol::attr("normalized")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ifft");
    
        auto result = at::ifft(std::move(peek(stack, 0, 1)), signal_ndim, normalized);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::index(Tensor self, Tensor[] indices) -> Tensor",
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("index");
    
        auto result = at::index(std::move(peek(stack, 0, varargs_length)), peekSlice(stack, 1, varargs_length - 1, varargs_length));
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::index_put(Tensor self, Tensor[] indices, Tensor values) -> Tensor",
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("index_put");
    
        auto result = at::index_put(std::move(peek(stack, 0, varargs_length)), peekSlice(stack, 1, varargs_length - 2, varargs_length), std::move(peek(stack, 1, 2)));
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::index_select(Tensor self, int dim, Tensor index) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("index_select");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto result = at::index_select(std::move(peek(stack, 0, 3)), dim, std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("index_select");
    
        auto result = at::index_select(std::move(peek(stack, 0, 2)), dim, std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::inverse(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("inverse");
    
        auto result = at::inverse(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::irfft(Tensor self, int signal_ndim, bool normalized=False, bool onesided=True, int[] signal_sizes=[]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("irfft");
        auto signal_ndim = tensor_as<int64_t>(std::move(peek(stack, 1, 5)));
        auto normalized = tensor_as<bool>(std::move(peek(stack, 2, 5)));
        auto onesided = tensor_as<bool>(std::move(peek(stack, 3, 5)));
        auto signal_sizes = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        auto result = at::irfft(std::move(peek(stack, 0, 5)), signal_ndim, normalized, onesided, signal_sizes);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto signal_ndim = int64_t(node->i(Symbol::attr("signal_ndim")));
      auto normalized = bool(node->i(Symbol::attr("normalized")));
      auto onesided = bool(node->i(Symbol::attr("onesided")));
      auto signal_sizes = std::vector<int64_t>(node->is(Symbol::attr("signal_sizes")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("irfft");
    
        auto result = at::irfft(std::move(peek(stack, 0, 1)), signal_ndim, normalized, onesided, signal_sizes);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::is_coalesced(Tensor self) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("is_coalesced");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).is_coalesced();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::is_contiguous(Tensor self) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("is_contiguous");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).is_contiguous();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::is_cuda(Tensor self) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("is_cuda");
    
        auto result = at::is_cuda(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::is_distributed(Tensor self) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("is_distributed");
    
        auto result = at::is_distributed(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::is_floating_point(Tensor self) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("is_floating_point");
    
        auto result = at::is_floating_point(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::is_nonzero(Tensor self) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("is_nonzero");
    
        auto result = at::is_nonzero(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::is_same_size(Tensor self, Tensor other) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("is_same_size");
    
        auto result = at::is_same_size(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::is_set_to(Tensor self, Tensor tensor) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("is_set_to");
    
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).is_set_to(std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::is_signed(Tensor self) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("is_signed");
    
        auto result = at::is_signed(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::is_sparse(Tensor self) -> bool",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("is_sparse");
    
        auto result = at::is_sparse(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::isclose(Tensor self, Tensor other, float rtol=1e-05, float atol=1e-08, bool equal_nan=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("isclose");
        auto rtol = tensor_as<double>(std::move(peek(stack, 2, 5)));
        auto atol = tensor_as<double>(std::move(peek(stack, 3, 5)));
        auto equal_nan = tensor_as<bool>(std::move(peek(stack, 4, 5)));
        auto result = at::isclose(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), rtol, atol, equal_nan);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto rtol = double(node->f(Symbol::attr("rtol")));
      auto atol = double(node->f(Symbol::attr("atol")));
      auto equal_nan = bool(node->i(Symbol::attr("equal_nan")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("isclose");
    
        auto result = at::isclose(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), rtol, atol, equal_nan);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::kl_div(Tensor self, Tensor target, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("kl_div");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::kl_div(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("kl_div");
    
        auto result = at::kl_div(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::kl_div_backward(Tensor grad_output, Tensor self, Tensor target, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("kl_div_backward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::kl_div_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), reduction);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("kl_div_backward");
    
        auto result = at::kl_div_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::kl_div_forward(Tensor self, Tensor target, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("kl_div_forward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::kl_div_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("kl_div_forward");
    
        auto result = at::kl_div_forward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::kthvalue(Tensor self, int k, int dim=-1, bool keepdim=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("kthvalue");
        auto k = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::kthvalue(std::move(peek(stack, 0, 4)), k, dim, keepdim);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto k = int64_t(node->i(Symbol::attr("k")));
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("kthvalue");
    
        auto result = at::kthvalue(std::move(peek(stack, 0, 1)), k, dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::l1_loss(Tensor self, Tensor target, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("l1_loss");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::l1_loss(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("l1_loss");
    
        auto result = at::l1_loss(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::l1_loss_backward(Tensor grad_output, Tensor self, Tensor target, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("l1_loss_backward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::l1_loss_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), reduction);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("l1_loss_backward");
    
        auto result = at::l1_loss_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::l1_loss_forward(Tensor self, Tensor target, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("l1_loss_forward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::l1_loss_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("l1_loss_forward");
    
        auto result = at::l1_loss_forward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::layer_norm(Tensor input, int[] normalized_shape, Tensor? weight=None, Tensor? bias=None, float eps=1e-05, bool cudnn_enable=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("layer_norm");
        auto normalized_shape = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto eps = tensor_as<double>(std::move(peek(stack, 4, 6)));
        auto cudnn_enable = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::layer_norm(std::move(peek(stack, 0, 6)), normalized_shape, std::move(peek(stack, 2, 6)), std::move(peek(stack, 3, 6)), eps, cudnn_enable);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto normalized_shape = std::vector<int64_t>(node->is(Symbol::attr("normalized_shape")));
      auto eps = double(node->f(Symbol::attr("eps")));
      auto cudnn_enable = bool(node->i(Symbol::attr("cudnn_enable")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("layer_norm");
    
        auto result = at::layer_norm(std::move(peek(stack, 0, 3)), normalized_shape, std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), eps, cudnn_enable);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::le(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("le");
    
        auto result = at::le(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::le(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("le");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::le(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("le");
    
        auto result = at::le(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::leaky_relu(Tensor self, Scalar negative_slope=0.01) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("leaky_relu");
        auto negative_slope = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::leaky_relu(std::move(peek(stack, 0, 2)), negative_slope);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto negative_slope = Scalar(node->t(Symbol::attr("negative_slope")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("leaky_relu");
    
        auto result = at::leaky_relu(std::move(peek(stack, 0, 1)), negative_slope);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::leaky_relu_backward(Tensor grad_output, Tensor self, Scalar negative_slope) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("leaky_relu_backward");
        auto negative_slope = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::leaky_relu_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), negative_slope);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto negative_slope = Scalar(node->t(Symbol::attr("negative_slope")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("leaky_relu_backward");
    
        auto result = at::leaky_relu_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), negative_slope);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::leaky_relu_forward(Tensor self, Scalar negative_slope) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("leaky_relu_forward");
        auto negative_slope = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::leaky_relu_forward(std::move(peek(stack, 0, 2)), negative_slope);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto negative_slope = Scalar(node->t(Symbol::attr("negative_slope")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("leaky_relu_forward");
    
        auto result = at::leaky_relu_forward(std::move(peek(stack, 0, 1)), negative_slope);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::lerp(Tensor self, Tensor end, Scalar weight) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("lerp");
        auto weight = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::lerp(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), weight);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto weight = Scalar(node->t(Symbol::attr("weight")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("lerp");
    
        auto result = at::lerp(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), weight);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::lgamma(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("lgamma");
    
        auto result = at::lgamma(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::linspace(Scalar start, Scalar end, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("linspace");
        auto start = tensor_as<Scalar>(std::move(peek(stack, 0, 5)));
        auto end = tensor_as<Scalar>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::linspace(start, end, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto start = Scalar(node->t(Symbol::attr("start")));
      auto end = Scalar(node->t(Symbol::attr("end")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("linspace");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::linspace(start, end, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::linspace(Scalar start, Scalar end, int steps, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("linspace");
        auto start = tensor_as<Scalar>(std::move(peek(stack, 0, 6)));
        auto end = tensor_as<Scalar>(std::move(peek(stack, 1, 6)));
        auto steps = tensor_as<int64_t>(std::move(peek(stack, 2, 6)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 3, 6)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 4, 6)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::linspace(start, end, steps, options);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto start = Scalar(node->t(Symbol::attr("start")));
      auto end = Scalar(node->t(Symbol::attr("end")));
      auto steps = int64_t(node->i(Symbol::attr("steps")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("linspace");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::linspace(start, end, steps, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::log(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log");
    
        auto result = at::log(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::log10(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log10");
    
        auto result = at::log10(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::log1p(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log1p");
    
        auto result = at::log1p(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::log2(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log2");
    
        auto result = at::log2(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::log_sigmoid(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log_sigmoid");
    
        auto result = at::log_sigmoid(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::log_sigmoid_backward(Tensor grad_output, Tensor self, Tensor buffer) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log_sigmoid_backward");
    
        auto result = at::log_sigmoid_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::log_sigmoid_forward(Tensor self) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log_sigmoid_forward");
    
        auto result = at::log_sigmoid_forward(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::log_softmax(Tensor self, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log_softmax");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::log_softmax(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log_softmax");
    
        auto result = at::log_softmax(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::log_softmax_backward_data(Tensor grad_output, Tensor output, int dim, Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log_softmax_backward_data");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto result = at::log_softmax_backward_data(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), dim, std::move(peek(stack, 3, 4)));
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("log_softmax_backward_data");
    
        auto result = at::log_softmax_backward_data(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), dim, std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::logdet(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("logdet");
    
        auto result = at::logdet(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::logspace(Scalar start, Scalar end, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("logspace");
        auto start = tensor_as<Scalar>(std::move(peek(stack, 0, 5)));
        auto end = tensor_as<Scalar>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::logspace(start, end, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto start = Scalar(node->t(Symbol::attr("start")));
      auto end = Scalar(node->t(Symbol::attr("end")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("logspace");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::logspace(start, end, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::logspace(Scalar start, Scalar end, int steps, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("logspace");
        auto start = tensor_as<Scalar>(std::move(peek(stack, 0, 6)));
        auto end = tensor_as<Scalar>(std::move(peek(stack, 1, 6)));
        auto steps = tensor_as<int64_t>(std::move(peek(stack, 2, 6)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 3, 6)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 4, 6)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::logspace(start, end, steps, options);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto start = Scalar(node->t(Symbol::attr("start")));
      auto end = Scalar(node->t(Symbol::attr("end")));
      auto steps = int64_t(node->i(Symbol::attr("steps")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("logspace");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::logspace(start, end, steps, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::logsumexp(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("logsumexp");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::logsumexp(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("logsumexp");
    
        auto result = at::logsumexp(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::lt(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("lt");
    
        auto result = at::lt(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::lt(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("lt");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::lt(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("lt");
    
        auto result = at::lt(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::margin_ranking_loss(Tensor input1, Tensor input2, Tensor target, float margin=0.0, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("margin_ranking_loss");
        auto margin = tensor_as<double>(std::move(peek(stack, 3, 5)));
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::margin_ranking_loss(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), margin, reduction);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto margin = double(node->f(Symbol::attr("margin")));
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("margin_ranking_loss");
    
        auto result = at::margin_ranking_loss(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), margin, reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::masked_select(Tensor self, Tensor mask) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("masked_select");
    
        auto result = at::masked_select(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::matmul(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("matmul");
    
        auto result = at::matmul(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max");
    
        auto result = at::max(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max");
    
        auto result = at::max(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max(Tensor self, int dim, bool keepdim=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::max(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max");
    
        auto result = at::max(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_pool1d(Tensor self, int[1] kernel_size, int[1] stride=[], int[1] padding=0, int[1] dilation=1, bool ceil_mode=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool1d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::max_pool1d(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool1d");
    
        auto result = at::max_pool1d(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_pool1d_with_indices(Tensor self, int[1] kernel_size, int[1] stride=[], int[1] padding=0, int[1] dilation=1, bool ceil_mode=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool1d_with_indices");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::max_pool1d_with_indices(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool1d_with_indices");
    
        auto result = at::max_pool1d_with_indices(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_pool2d(Tensor self, int[1] kernel_size, int[1] stride=[], int[1] padding=0, int[1] dilation=1, bool ceil_mode=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool2d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::max_pool2d(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool2d");
    
        auto result = at::max_pool2d(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_pool2d_with_indices(Tensor self, int[2] kernel_size, int[2] stride=[], int[2] padding=0, int[2] dilation=1, bool ceil_mode=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool2d_with_indices");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::max_pool2d_with_indices(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool2d_with_indices");
    
        auto result = at::max_pool2d_with_indices(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_pool2d_with_indices_backward(Tensor grad_output, Tensor self, int[2] kernel_size, int[2] stride, int[2] padding, int[2] dilation, bool ceil_mode, Tensor indices) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool2d_with_indices_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 8)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 6, 8)));
        auto result = at::max_pool2d_with_indices_backward(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), kernel_size, stride, padding, dilation, ceil_mode, std::move(peek(stack, 7, 8)));
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool2d_with_indices_backward");
    
        auto result = at::max_pool2d_with_indices_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, stride, padding, dilation, ceil_mode, std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_pool2d_with_indices_forward(Tensor self, int[2] kernel_size, int[2] stride, int[2] padding, int[2] dilation, bool ceil_mode) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool2d_with_indices_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::max_pool2d_with_indices_forward(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool2d_with_indices_forward");
    
        auto result = at::max_pool2d_with_indices_forward(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_pool3d(Tensor self, int[1] kernel_size, int[1] stride=[], int[1] padding=0, int[1] dilation=1, bool ceil_mode=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool3d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::max_pool3d(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool3d");
    
        auto result = at::max_pool3d(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_pool3d_with_indices(Tensor self, int[3] kernel_size, int[3] stride=[], int[3] padding=0, int[3] dilation=1, bool ceil_mode=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool3d_with_indices");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::max_pool3d_with_indices(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool3d_with_indices");
    
        auto result = at::max_pool3d_with_indices(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_pool3d_with_indices_backward(Tensor grad_output, Tensor self, int[3] kernel_size, int[3] stride, int[3] padding, int[3] dilation, bool ceil_mode, Tensor indices) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool3d_with_indices_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 8)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 6, 8)));
        auto result = at::max_pool3d_with_indices_backward(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), kernel_size, stride, padding, dilation, ceil_mode, std::move(peek(stack, 7, 8)));
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool3d_with_indices_backward");
    
        auto result = at::max_pool3d_with_indices_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, stride, padding, dilation, ceil_mode, std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_pool3d_with_indices_forward(Tensor self, int[3] kernel_size, int[3] stride, int[3] padding, int[3] dilation, bool ceil_mode) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool3d_with_indices_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 1, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto ceil_mode = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::max_pool3d_with_indices_forward(std::move(peek(stack, 0, 6)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto ceil_mode = bool(node->i(Symbol::attr("ceil_mode")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_pool3d_with_indices_forward");
    
        auto result = at::max_pool3d_with_indices_forward(std::move(peek(stack, 0, 1)), kernel_size, stride, padding, dilation, ceil_mode);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_unpool2d(Tensor self, Tensor indices, int[2] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool2d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::max_unpool2d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), output_size);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool2d");
    
        auto result = at::max_unpool2d(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_unpool2d_backward(Tensor grad_output, Tensor self, Tensor indices, int[2] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool2d_backward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        auto result = at::max_unpool2d_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), output_size);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool2d_backward");
    
        auto result = at::max_unpool2d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), output_size);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_unpool2d_forward(Tensor self, Tensor indices, int[2] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool2d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::max_unpool2d_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), output_size);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool2d_forward");
    
        auto result = at::max_unpool2d_forward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_unpool3d(Tensor self, Tensor indices, int[3] output_size, int[3] stride, int[3] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool3d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 2, 5)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 5)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        auto result = at::max_unpool3d(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), output_size, stride, padding);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool3d");
    
        auto result = at::max_unpool3d(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), output_size, stride, padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_unpool3d_backward(Tensor grad_output, Tensor self, Tensor indices, int[3] output_size, int[3] stride, int[3] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool3d_backward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        auto result = at::max_unpool3d_backward(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), std::move(peek(stack, 2, 6)), output_size, stride, padding);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool3d_backward");
    
        auto result = at::max_unpool3d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), output_size, stride, padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_unpool3d_forward(Tensor self, Tensor indices, int[3] output_size, int[3] stride, int[3] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool3d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 2, 5)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 3, 5)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        auto result = at::max_unpool3d_forward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), output_size, stride, padding);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_unpool3d_forward");
    
        auto result = at::max_unpool3d_forward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), output_size, stride, padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::max_values(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_values");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::max_values(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("max_values");
    
        auto result = at::max_values(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mean(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mean");
    
        auto result = at::mean(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mean(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mean");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::mean(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mean");
    
        auto result = at::mean(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::median(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("median");
    
        auto result = at::median(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::median(Tensor self, int dim, bool keepdim=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("median");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::median(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("median");
    
        auto result = at::median(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::meshgrid(Tensor[] tensors) -> Tensor[]",
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("meshgrid");
    
        auto result = at::meshgrid(peekSlice(stack, 0, varargs_length - 0, varargs_length));
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::min(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("min");
    
        auto result = at::min(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::min(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("min");
    
        auto result = at::min(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::min(Tensor self, int dim, bool keepdim=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("min");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::min(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("min");
    
        auto result = at::min(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::min_values(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("min_values");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::min_values(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("min_values");
    
        auto result = at::min_values(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mkldnn_convolution(Tensor self, Tensor weight, Tensor? bias, int[] padding, int[] stride, int[] dilation) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mkldnn_convolution");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        auto result = at::mkldnn_convolution(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), std::move(peek(stack, 2, 6)), padding, stride, dilation);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mkldnn_convolution");
    
        auto result = at::mkldnn_convolution(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), padding, stride, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mkldnn_convolution_backward(Tensor self, Tensor grad_output, Tensor weight, int[] padding, int[] stride, int[] dilation, bool[3] output_mask) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mkldnn_convolution_backward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 6, 7)));
        auto result = at::mkldnn_convolution_backward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), padding, stride, dilation, output_mask);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mkldnn_convolution_backward");
    
        auto result = at::mkldnn_convolution_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), padding, stride, dilation, output_mask);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mkldnn_convolution_backward_input(int[] self_size, Tensor grad_output, Tensor weight, int[] padding, int[] stride, int[] dilation, bool bias_defined) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mkldnn_convolution_backward_input");
        auto self_size = tensor_as<IntList>(std::move(peek(stack, 0, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto bias_defined = tensor_as<bool>(std::move(peek(stack, 6, 7)));
        auto result = at::mkldnn_convolution_backward_input(self_size, std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), padding, stride, dilation, bias_defined);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto self_size = std::vector<int64_t>(node->is(Symbol::attr("self_size")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto bias_defined = bool(node->i(Symbol::attr("bias_defined")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mkldnn_convolution_backward_input");
    
        auto result = at::mkldnn_convolution_backward_input(self_size, std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding, stride, dilation, bias_defined);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mkldnn_convolution_backward_weights(int[] weight_size, Tensor grad_output, Tensor self, int[] padding, int[] stride, int[] dilation, bool bias_defined) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mkldnn_convolution_backward_weights");
        auto weight_size = tensor_as<IntList>(std::move(peek(stack, 0, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 3, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto bias_defined = tensor_as<bool>(std::move(peek(stack, 6, 7)));
        auto result = at::mkldnn_convolution_backward_weights(weight_size, std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), padding, stride, dilation, bias_defined);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto weight_size = std::vector<int64_t>(node->is(Symbol::attr("weight_size")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto bias_defined = bool(node->i(Symbol::attr("bias_defined")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mkldnn_convolution_backward_weights");
    
        auto result = at::mkldnn_convolution_backward_weights(weight_size, std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding, stride, dilation, bias_defined);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mm(Tensor self, Tensor mat2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mm");
    
        auto result = at::mm(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mode(Tensor self, int dim=-1, bool keepdim=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mode");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::mode(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mode");
    
        auto result = at::mode(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mse_loss(Tensor self, Tensor target, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mse_loss");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::mse_loss(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mse_loss");
    
        auto result = at::mse_loss(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mse_loss_backward(Tensor grad_output, Tensor self, Tensor target, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mse_loss_backward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::mse_loss_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), reduction);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mse_loss_backward");
    
        auto result = at::mse_loss_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mse_loss_forward(Tensor self, Tensor target, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mse_loss_forward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::mse_loss_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mse_loss_forward");
    
        auto result = at::mse_loss_forward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mul(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mul");
    
        auto result = at::mul(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mul(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mul");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::mul(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mul");
    
        auto result = at::mul(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::multi_margin_loss(Tensor self, Tensor target, Scalar p=1, Scalar margin=1, Tensor weight=None, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multi_margin_loss");
        auto p = tensor_as<Scalar>(std::move(peek(stack, 2, 6)));
        auto margin = tensor_as<Scalar>(std::move(peek(stack, 3, 6)));
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 5, 6)));
        auto result = at::multi_margin_loss(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), p, margin, std::move(peek(stack, 4, 6)), reduction);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = Scalar(node->t(Symbol::attr("p")));
      auto margin = Scalar(node->t(Symbol::attr("margin")));
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multi_margin_loss");
    
        auto result = at::multi_margin_loss(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), p, margin, std::move(peek(stack, 2, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::multi_margin_loss_backward(Tensor grad_output, Tensor self, Tensor target, Scalar p, Scalar margin, Tensor? weight, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multi_margin_loss_backward");
        auto p = tensor_as<Scalar>(std::move(peek(stack, 3, 7)));
        auto margin = tensor_as<Scalar>(std::move(peek(stack, 4, 7)));
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 6, 7)));
        auto result = at::multi_margin_loss_backward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), p, margin, std::move(peek(stack, 5, 7)), reduction);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = Scalar(node->t(Symbol::attr("p")));
      auto margin = Scalar(node->t(Symbol::attr("margin")));
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multi_margin_loss_backward");
    
        auto result = at::multi_margin_loss_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), p, margin, std::move(peek(stack, 3, 4)), reduction);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::multi_margin_loss_forward(Tensor self, Tensor target, Scalar p, Scalar margin, Tensor? weight, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multi_margin_loss_forward");
        auto p = tensor_as<Scalar>(std::move(peek(stack, 2, 6)));
        auto margin = tensor_as<Scalar>(std::move(peek(stack, 3, 6)));
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 5, 6)));
        auto result = at::multi_margin_loss_forward(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), p, margin, std::move(peek(stack, 4, 6)), reduction);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = Scalar(node->t(Symbol::attr("p")));
      auto margin = Scalar(node->t(Symbol::attr("margin")));
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multi_margin_loss_forward");
    
        auto result = at::multi_margin_loss_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), p, margin, std::move(peek(stack, 2, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::multilabel_margin_loss(Tensor self, Tensor target, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multilabel_margin_loss");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::multilabel_margin_loss(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multilabel_margin_loss");
    
        auto result = at::multilabel_margin_loss(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::multilabel_margin_loss_backward(Tensor grad_output, Tensor self, Tensor target, int reduction, Tensor is_target) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multilabel_margin_loss_backward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto result = at::multilabel_margin_loss_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), reduction, std::move(peek(stack, 4, 5)));
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multilabel_margin_loss_backward");
    
        auto result = at::multilabel_margin_loss_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), reduction, std::move(peek(stack, 3, 4)));
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::multilabel_margin_loss_forward(Tensor self, Tensor target, int reduction) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multilabel_margin_loss_forward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::multilabel_margin_loss_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multilabel_margin_loss_forward");
    
        auto result = at::multilabel_margin_loss_forward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::multinomial(Tensor self, int num_samples, bool replacement=False, *, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multinomial");
        auto num_samples = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto replacement = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::multinomial(std::move(peek(stack, 0, 3)), num_samples, replacement, nullptr);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto num_samples = int64_t(node->i(Symbol::attr("num_samples")));
      auto replacement = bool(node->i(Symbol::attr("replacement")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("multinomial");
    
        auto result = at::multinomial(std::move(peek(stack, 0, 1)), num_samples, replacement, nullptr);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::mv(Tensor self, Tensor vec) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("mv");
    
        auto result = at::mv(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::narrow(Tensor self, int dim, int start, int length) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("narrow");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto start = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto length = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::narrow(std::move(peek(stack, 0, 4)), dim, start, length);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto start = int64_t(node->i(Symbol::attr("start")));
      auto length = int64_t(node->i(Symbol::attr("length")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("narrow");
    
        auto result = at::narrow(std::move(peek(stack, 0, 1)), dim, start, length);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::native_clone(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("native_clone");
    
        auto result = at::native_clone(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::native_div(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("native_div");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::native_div(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("native_div");
    
        auto result = at::native_div(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::native_get_device(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("native_get_device");
    
        auto result = at::native_get_device(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::native_mul(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("native_mul");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::native_mul(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("native_mul");
    
        auto result = at::native_mul(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::native_norm(Tensor self, Scalar p=2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("native_norm");
        auto p = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::native_norm(std::move(peek(stack, 0, 2)), p);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = Scalar(node->t(Symbol::attr("p")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("native_norm");
    
        auto result = at::native_norm(std::move(peek(stack, 0, 1)), p);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::native_pow(Tensor self, Scalar exponent) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("native_pow");
        auto exponent = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::native_pow(std::move(peek(stack, 0, 2)), exponent);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto exponent = Scalar(node->t(Symbol::attr("exponent")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("native_pow");
    
        auto result = at::native_pow(std::move(peek(stack, 0, 1)), exponent);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ne(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ne");
    
        auto result = at::ne(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ne(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ne");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::ne(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ne");
    
        auto result = at::ne(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::neg(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("neg");
    
        auto result = at::neg(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::nll_loss(Tensor self, Tensor target, Tensor weight=None, int reduction=ElementwiseMean, int ignore_index=-100) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto ignore_index = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::nll_loss(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), reduction, ignore_index);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      auto ignore_index = int64_t(node->i(Symbol::attr("ignore_index")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss");
    
        auto result = at::nll_loss(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction, ignore_index);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::nll_loss2d(Tensor self, Tensor target, Tensor weight=None, int reduction=ElementwiseMean, int ignore_index=-100) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss2d");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto ignore_index = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::nll_loss2d(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), reduction, ignore_index);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      auto ignore_index = int64_t(node->i(Symbol::attr("ignore_index")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss2d");
    
        auto result = at::nll_loss2d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction, ignore_index);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::nll_loss2d_backward(Tensor grad_output, Tensor self, Tensor target, Tensor? weight, int reduction, int ignore_index, Tensor total_weight) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss2d_backward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 4, 7)));
        auto ignore_index = tensor_as<int64_t>(std::move(peek(stack, 5, 7)));
        auto result = at::nll_loss2d_backward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), std::move(peek(stack, 3, 7)), reduction, ignore_index, std::move(peek(stack, 6, 7)));
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      auto ignore_index = int64_t(node->i(Symbol::attr("ignore_index")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss2d_backward");
    
        auto result = at::nll_loss2d_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), std::move(peek(stack, 3, 5)), reduction, ignore_index, std::move(peek(stack, 4, 5)));
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::nll_loss2d_forward(Tensor self, Tensor target, Tensor? weight, int reduction, int ignore_index) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss2d_forward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto ignore_index = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::nll_loss2d_forward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), reduction, ignore_index);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      auto ignore_index = int64_t(node->i(Symbol::attr("ignore_index")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss2d_forward");
    
        auto result = at::nll_loss2d_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction, ignore_index);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::nll_loss_backward(Tensor grad_output, Tensor self, Tensor target, Tensor? weight, int reduction, int ignore_index, Tensor total_weight) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss_backward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 4, 7)));
        auto ignore_index = tensor_as<int64_t>(std::move(peek(stack, 5, 7)));
        auto result = at::nll_loss_backward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), std::move(peek(stack, 3, 7)), reduction, ignore_index, std::move(peek(stack, 6, 7)));
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      auto ignore_index = int64_t(node->i(Symbol::attr("ignore_index")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss_backward");
    
        auto result = at::nll_loss_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), std::move(peek(stack, 3, 5)), reduction, ignore_index, std::move(peek(stack, 4, 5)));
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::nll_loss_forward(Tensor self, Tensor target, Tensor? weight, int reduction, int ignore_index) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss_forward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto ignore_index = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::nll_loss_forward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), reduction, ignore_index);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      auto ignore_index = int64_t(node->i(Symbol::attr("ignore_index")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nll_loss_forward");
    
        auto result = at::nll_loss_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction, ignore_index);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::nonzero(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("nonzero");
    
        auto result = at::nonzero(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::norm(Tensor self, Scalar p=2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("norm");
        auto p = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::norm(std::move(peek(stack, 0, 2)), p);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = Scalar(node->t(Symbol::attr("p")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("norm");
    
        auto result = at::norm(std::move(peek(stack, 0, 1)), p);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::norm(Tensor self, Scalar p, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("norm");
        auto p = tensor_as<Scalar>(std::move(peek(stack, 1, 4)));
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::norm(std::move(peek(stack, 0, 4)), p, dim, keepdim);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = Scalar(node->t(Symbol::attr("p")));
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("norm");
    
        auto result = at::norm(std::move(peek(stack, 0, 1)), p, dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::normal(Tensor mean, Tensor std, *, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("normal");
    
        auto result = at::normal(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), nullptr);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::normal(float mean, Tensor std, *, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("normal");
        auto mean = tensor_as<double>(std::move(peek(stack, 0, 2)));
        auto result = at::normal(mean, std::move(peek(stack, 1, 2)), nullptr);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto mean = double(node->f(Symbol::attr("mean")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("normal");
    
        auto result = at::normal(mean, std::move(peek(stack, 0, 1)), nullptr);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::normal(Tensor mean, float std=1, *, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("normal");
        auto std = tensor_as<double>(std::move(peek(stack, 1, 2)));
        auto result = at::normal(std::move(peek(stack, 0, 2)), std, nullptr);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto std = double(node->f(Symbol::attr("std")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("normal");
    
        auto result = at::normal(std::move(peek(stack, 0, 1)), std, nullptr);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::numel(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("numel");
    
        auto result = at::numel(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ones(int[] size, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ones");
        auto size = tensor_as<IntList>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::ones(size, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ones");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::ones(size, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ones_like(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ones_like");
    
        auto result = at::ones_like(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ones_like(Tensor self, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ones_like");
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::ones_like(std::move(peek(stack, 0, 4)), options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ones_like");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::ones_like(std::move(peek(stack, 0, 1)), options);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::orgqr(Tensor self, Tensor input2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("orgqr");
    
        auto result = at::orgqr(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::ormqr(Tensor self, Tensor input2, Tensor input3, bool left=True, bool transpose=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ormqr");
        auto left = tensor_as<bool>(std::move(peek(stack, 3, 5)));
        auto transpose = tensor_as<bool>(std::move(peek(stack, 4, 5)));
        auto result = at::ormqr(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), left, transpose);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto left = bool(node->i(Symbol::attr("left")));
      auto transpose = bool(node->i(Symbol::attr("transpose")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("ormqr");
    
        auto result = at::ormqr(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), left, transpose);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::pairwise_distance(Tensor x1, Tensor x2, float p=2, float eps=1e-06, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pairwise_distance");
        auto p = tensor_as<double>(std::move(peek(stack, 2, 5)));
        auto eps = tensor_as<double>(std::move(peek(stack, 3, 5)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 4, 5)));
        auto result = at::pairwise_distance(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), p, eps, keepdim);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = double(node->f(Symbol::attr("p")));
      auto eps = double(node->f(Symbol::attr("eps")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pairwise_distance");
    
        auto result = at::pairwise_distance(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), p, eps, keepdim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::permute(Tensor self, int[] dims) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("permute");
        auto dims = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).permute(dims);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dims = std::vector<int64_t>(node->is(Symbol::attr("dims")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("permute");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).permute(dims);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::pin_memory(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pin_memory");
    
        auto result = at::pin_memory(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::pinverse(Tensor self, float rcond=1e-15) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pinverse");
        auto rcond = tensor_as<double>(std::move(peek(stack, 1, 2)));
        auto result = at::pinverse(std::move(peek(stack, 0, 2)), rcond);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto rcond = double(node->f(Symbol::attr("rcond")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pinverse");
    
        auto result = at::pinverse(std::move(peek(stack, 0, 1)), rcond);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::poisson(Tensor self, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("poisson");
    
        auto result = at::poisson(std::move(peek(stack, 0, 1)), nullptr);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::polygamma(int n, Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("polygamma");
        auto n = tensor_as<int64_t>(std::move(peek(stack, 0, 2)));
        auto result = at::polygamma(n, std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto n = int64_t(node->i(Symbol::attr("n")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("polygamma");
    
        auto result = at::polygamma(n, std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::potrf(Tensor self, bool upper=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("potrf");
        auto upper = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::potrf(std::move(peek(stack, 0, 2)), upper);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto upper = bool(node->i(Symbol::attr("upper")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("potrf");
    
        auto result = at::potrf(std::move(peek(stack, 0, 1)), upper);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::potri(Tensor self, bool upper=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("potri");
        auto upper = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::potri(std::move(peek(stack, 0, 2)), upper);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto upper = bool(node->i(Symbol::attr("upper")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("potri");
    
        auto result = at::potri(std::move(peek(stack, 0, 1)), upper);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::potrs(Tensor self, Tensor input2, bool upper=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("potrs");
        auto upper = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::potrs(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), upper);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto upper = bool(node->i(Symbol::attr("upper")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("potrs");
    
        auto result = at::potrs(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), upper);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::pow(Tensor self, Tensor exponent) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pow");
    
        auto result = at::pow(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::pow(Scalar base, Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pow");
        auto base = tensor_as<Scalar>(std::move(peek(stack, 0, 2)));
        auto result = at::pow(base, std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto base = Scalar(node->t(Symbol::attr("base")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pow");
    
        auto result = at::pow(base, std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::pow(Tensor self, Scalar exponent) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pow");
        auto exponent = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::pow(std::move(peek(stack, 0, 2)), exponent);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto exponent = Scalar(node->t(Symbol::attr("exponent")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pow");
    
        auto result = at::pow(std::move(peek(stack, 0, 1)), exponent);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::prelu(Tensor self, Tensor weight) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("prelu");
    
        auto result = at::prelu(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::prelu_backward(Tensor grad_output, Tensor self, Tensor weight, bool[2] output_mask=[True, True]) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("prelu_backward");
        auto output_mask = tensor_as<std::array<bool,2>>(std::move(peek(stack, 3, 4)));
        auto result = at::prelu_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), output_mask);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_mask = as_bool_array<2>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("prelu_backward");
    
        auto result = at::prelu_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), output_mask);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::prelu_forward(Tensor self, Tensor weight) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("prelu_forward");
    
        auto result = at::prelu_forward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::prod(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("prod");
    
        auto result = at::prod(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::prod(Tensor self, int dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("prod");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::prod(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("prod");
    
        auto result = at::prod(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::pstrf(Tensor self, bool upper=True, Scalar tol=-1) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pstrf");
        auto upper = tensor_as<bool>(std::move(peek(stack, 1, 3)));
        auto tol = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::pstrf(std::move(peek(stack, 0, 3)), upper, tol);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto upper = bool(node->i(Symbol::attr("upper")));
      auto tol = Scalar(node->t(Symbol::attr("tol")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("pstrf");
    
        auto result = at::pstrf(std::move(peek(stack, 0, 1)), upper, tol);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::qr(Tensor self) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("qr");
    
        auto result = at::qr(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::rand(int[] size, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rand");
        auto size = tensor_as<IntList>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::rand(size, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rand");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::rand(size, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::rand_like(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rand_like");
    
        auto result = at::rand_like(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::rand_like(Tensor self, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rand_like");
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::rand_like(std::move(peek(stack, 0, 4)), options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rand_like");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::rand_like(std::move(peek(stack, 0, 1)), options);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::randint(int high, int[] size, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint");
        auto high = tensor_as<int64_t>(std::move(peek(stack, 0, 5)));
        auto size = tensor_as<IntList>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randint(high, size, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto high = int64_t(node->i(Symbol::attr("high")));
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randint(high, size, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::randint(int low, int high, int[] size, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint");
        auto low = tensor_as<int64_t>(std::move(peek(stack, 0, 6)));
        auto high = tensor_as<int64_t>(std::move(peek(stack, 1, 6)));
        auto size = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 3, 6)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 4, 6)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randint(low, high, size, options);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto low = int64_t(node->i(Symbol::attr("low")));
      auto high = int64_t(node->i(Symbol::attr("high")));
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randint(low, high, size, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::randint_like(Tensor self, int high) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint_like");
        auto high = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::randint_like(std::move(peek(stack, 0, 2)), high);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto high = int64_t(node->i(Symbol::attr("high")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint_like");
    
        auto result = at::randint_like(std::move(peek(stack, 0, 1)), high);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::randint_like(Tensor self, int low, int high) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint_like");
        auto low = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto high = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::randint_like(std::move(peek(stack, 0, 3)), low, high);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto low = int64_t(node->i(Symbol::attr("low")));
      auto high = int64_t(node->i(Symbol::attr("high")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint_like");
    
        auto result = at::randint_like(std::move(peek(stack, 0, 1)), low, high);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::randint_like(Tensor self, int high, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint_like");
        auto high = tensor_as<int64_t>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randint_like(std::move(peek(stack, 0, 5)), high, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto high = int64_t(node->i(Symbol::attr("high")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint_like");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randint_like(std::move(peek(stack, 0, 1)), high, options);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::randint_like(Tensor self, int low, int high, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint_like");
        auto low = tensor_as<int64_t>(std::move(peek(stack, 1, 6)));
        auto high = tensor_as<int64_t>(std::move(peek(stack, 2, 6)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 3, 6)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 4, 6)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randint_like(std::move(peek(stack, 0, 6)), low, high, options);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto low = int64_t(node->i(Symbol::attr("low")));
      auto high = int64_t(node->i(Symbol::attr("high")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randint_like");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randint_like(std::move(peek(stack, 0, 1)), low, high, options);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::randn(int[] size, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randn");
        auto size = tensor_as<IntList>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randn(size, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randn");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randn(size, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::randn_like(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randn_like");
    
        auto result = at::randn_like(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::randn_like(Tensor self, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randn_like");
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randn_like(std::move(peek(stack, 0, 4)), options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randn_like");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randn_like(std::move(peek(stack, 0, 1)), options);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::randperm(int n, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randperm");
        auto n = tensor_as<int64_t>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randperm(n, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto n = int64_t(node->i(Symbol::attr("n")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("randperm");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::randperm(n, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::range(Scalar start, Scalar end, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("range");
        auto start = tensor_as<Scalar>(std::move(peek(stack, 0, 5)));
        auto end = tensor_as<Scalar>(std::move(peek(stack, 1, 5)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 4, 5)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::range(start, end, options);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto start = Scalar(node->t(Symbol::attr("start")));
      auto end = Scalar(node->t(Symbol::attr("end")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("range");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::range(start, end, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::range(Scalar start, Scalar end, Scalar step, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("range");
        auto start = tensor_as<Scalar>(std::move(peek(stack, 0, 6)));
        auto end = tensor_as<Scalar>(std::move(peek(stack, 1, 6)));
        auto step = tensor_as<Scalar>(std::move(peek(stack, 2, 6)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 3, 6)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 4, 6)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::range(start, end, step, options);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto start = Scalar(node->t(Symbol::attr("start")));
      auto end = Scalar(node->t(Symbol::attr("end")));
      auto step = Scalar(node->t(Symbol::attr("step")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("range");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::range(start, end, step, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::reciprocal(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reciprocal");
    
        auto result = at::reciprocal(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::reflection_pad1d(Tensor self, int[2] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad1d");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::reflection_pad1d(std::move(peek(stack, 0, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad1d");
    
        auto result = at::reflection_pad1d(std::move(peek(stack, 0, 1)), padding);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::reflection_pad1d_backward(Tensor grad_output, Tensor self, int[2] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad1d_backward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::reflection_pad1d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad1d_backward");
    
        auto result = at::reflection_pad1d_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::reflection_pad1d_forward(Tensor self, int[2] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad1d_forward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::reflection_pad1d_forward(std::move(peek(stack, 0, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad1d_forward");
    
        auto result = at::reflection_pad1d_forward(std::move(peek(stack, 0, 1)), padding);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::reflection_pad2d(Tensor self, int[4] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad2d");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::reflection_pad2d(std::move(peek(stack, 0, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad2d");
    
        auto result = at::reflection_pad2d(std::move(peek(stack, 0, 1)), padding);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::reflection_pad2d_backward(Tensor grad_output, Tensor self, int[4] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad2d_backward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::reflection_pad2d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad2d_backward");
    
        auto result = at::reflection_pad2d_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::reflection_pad2d_forward(Tensor self, int[4] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad2d_forward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::reflection_pad2d_forward(std::move(peek(stack, 0, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reflection_pad2d_forward");
    
        auto result = at::reflection_pad2d_forward(std::move(peek(stack, 0, 1)), padding);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::relu(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("relu");
    
        auto result = at::relu(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::remainder(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("remainder");
    
        auto result = at::remainder(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::remainder(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("remainder");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::remainder(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("remainder");
    
        auto result = at::remainder(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::renorm(Tensor self, Scalar p, int dim, Scalar maxnorm) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("renorm");
        auto p = tensor_as<Scalar>(std::move(peek(stack, 1, 4)));
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto maxnorm = tensor_as<Scalar>(std::move(peek(stack, 3, 4)));
        auto result = at::renorm(std::move(peek(stack, 0, 4)), p, dim, maxnorm);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = Scalar(node->t(Symbol::attr("p")));
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto maxnorm = Scalar(node->t(Symbol::attr("maxnorm")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("renorm");
    
        auto result = at::renorm(std::move(peek(stack, 0, 1)), p, dim, maxnorm);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::repeat(Tensor self, int[] repeats) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("repeat");
        auto repeats = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).repeat(repeats);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto repeats = std::vector<int64_t>(node->is(Symbol::attr("repeats")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("repeat");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).repeat(repeats);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::replication_pad1d(Tensor self, int[2] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad1d");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::replication_pad1d(std::move(peek(stack, 0, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad1d");
    
        auto result = at::replication_pad1d(std::move(peek(stack, 0, 1)), padding);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::replication_pad1d_backward(Tensor grad_output, Tensor self, int[2] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad1d_backward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::replication_pad1d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad1d_backward");
    
        auto result = at::replication_pad1d_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::replication_pad1d_forward(Tensor self, int[2] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad1d_forward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::replication_pad1d_forward(std::move(peek(stack, 0, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad1d_forward");
    
        auto result = at::replication_pad1d_forward(std::move(peek(stack, 0, 1)), padding);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::replication_pad2d(Tensor self, int[4] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad2d");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::replication_pad2d(std::move(peek(stack, 0, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad2d");
    
        auto result = at::replication_pad2d(std::move(peek(stack, 0, 1)), padding);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::replication_pad2d_backward(Tensor grad_output, Tensor self, int[4] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad2d_backward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::replication_pad2d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad2d_backward");
    
        auto result = at::replication_pad2d_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::replication_pad2d_forward(Tensor self, int[4] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad2d_forward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::replication_pad2d_forward(std::move(peek(stack, 0, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad2d_forward");
    
        auto result = at::replication_pad2d_forward(std::move(peek(stack, 0, 1)), padding);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::replication_pad3d(Tensor self, int[6] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad3d");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::replication_pad3d(std::move(peek(stack, 0, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad3d");
    
        auto result = at::replication_pad3d(std::move(peek(stack, 0, 1)), padding);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::replication_pad3d_backward(Tensor grad_output, Tensor self, int[6] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad3d_backward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::replication_pad3d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad3d_backward");
    
        auto result = at::replication_pad3d_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::replication_pad3d_forward(Tensor self, int[6] padding) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad3d_forward");
        auto padding = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::replication_pad3d_forward(std::move(peek(stack, 0, 2)), padding);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("replication_pad3d_forward");
    
        auto result = at::replication_pad3d_forward(std::move(peek(stack, 0, 1)), padding);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::reshape(Tensor self, int[] shape) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reshape");
        auto shape = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::reshape(std::move(peek(stack, 0, 2)), shape);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto shape = std::vector<int64_t>(node->is(Symbol::attr("shape")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("reshape");
    
        auto result = at::reshape(std::move(peek(stack, 0, 1)), shape);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::rfft(Tensor self, int signal_ndim, bool normalized=False, bool onesided=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rfft");
        auto signal_ndim = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto normalized = tensor_as<bool>(std::move(peek(stack, 2, 4)));
        auto onesided = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::rfft(std::move(peek(stack, 0, 4)), signal_ndim, normalized, onesided);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto signal_ndim = int64_t(node->i(Symbol::attr("signal_ndim")));
      auto normalized = bool(node->i(Symbol::attr("normalized")));
      auto onesided = bool(node->i(Symbol::attr("onesided")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rfft");
    
        auto result = at::rfft(std::move(peek(stack, 0, 1)), signal_ndim, normalized, onesided);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::round(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("round");
    
        auto result = at::round(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::rrelu(Tensor self, Scalar lower=0.125, Scalar upper=0.3333333333333333, bool training=False, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rrelu");
        auto lower = tensor_as<Scalar>(std::move(peek(stack, 1, 4)));
        auto upper = tensor_as<Scalar>(std::move(peek(stack, 2, 4)));
        auto training = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::rrelu(std::move(peek(stack, 0, 4)), lower, upper, training, nullptr);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto lower = Scalar(node->t(Symbol::attr("lower")));
      auto upper = Scalar(node->t(Symbol::attr("upper")));
      auto training = bool(node->i(Symbol::attr("training")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rrelu");
    
        auto result = at::rrelu(std::move(peek(stack, 0, 1)), lower, upper, training, nullptr);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::rrelu_with_noise(Tensor self, Tensor noise, Scalar lower=0.125, Scalar upper=0.3333333333333333, bool training=False, Generator generator=None) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rrelu_with_noise");
        auto lower = tensor_as<Scalar>(std::move(peek(stack, 2, 5)));
        auto upper = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto training = tensor_as<bool>(std::move(peek(stack, 4, 5)));
        auto result = at::rrelu_with_noise(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), lower, upper, training, nullptr);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto lower = Scalar(node->t(Symbol::attr("lower")));
      auto upper = Scalar(node->t(Symbol::attr("upper")));
      auto training = bool(node->i(Symbol::attr("training")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rrelu_with_noise");
    
        auto result = at::rrelu_with_noise(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), lower, upper, training, nullptr);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::rrelu_with_noise_backward(Tensor grad_output, Tensor self, Tensor noise, Scalar lower, Scalar upper, bool training) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rrelu_with_noise_backward");
        auto lower = tensor_as<Scalar>(std::move(peek(stack, 3, 6)));
        auto upper = tensor_as<Scalar>(std::move(peek(stack, 4, 6)));
        auto training = tensor_as<bool>(std::move(peek(stack, 5, 6)));
        auto result = at::rrelu_with_noise_backward(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), std::move(peek(stack, 2, 6)), lower, upper, training);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto lower = Scalar(node->t(Symbol::attr("lower")));
      auto upper = Scalar(node->t(Symbol::attr("upper")));
      auto training = bool(node->i(Symbol::attr("training")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rrelu_with_noise_backward");
    
        auto result = at::rrelu_with_noise_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), lower, upper, training);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::rsqrt(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("rsqrt");
    
        auto result = at::rsqrt(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::s_native_add(Tensor self, Tensor other, *, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("s_native_add");
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::s_native_add(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("s_native_add");
    
        auto result = at::s_native_add(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), alpha);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::s_native_addmm(Tensor self, Tensor mat1, Tensor mat2, *, Scalar beta=1, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("s_native_addmm");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 4, 5)));
        auto result = at::s_native_addmm(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), beta, alpha);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("s_native_addmm");
    
        auto result = at::s_native_addmm(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), beta, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::s_native_mul(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("s_native_mul");
    
        auto result = at::s_native_mul(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::s_native_sub(Tensor self, Tensor other, *, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("s_native_sub");
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::s_native_sub(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("s_native_sub");
    
        auto result = at::s_native_sub(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), alpha);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::select(Tensor self, int dim, int index) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("select");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto index = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::select(std::move(peek(stack, 0, 3)), dim, index);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto index = int64_t(node->i(Symbol::attr("index")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("select");
    
        auto result = at::select(std::move(peek(stack, 0, 1)), dim, index);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::selu(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("selu");
    
        auto result = at::selu(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sigmoid(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sigmoid");
    
        auto result = at::sigmoid(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sign(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sign");
    
        auto result = at::sign(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sin(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sin");
    
        auto result = at::sin(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sinh(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sinh");
    
        auto result = at::sinh(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::size(Tensor self, int dim) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("size");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::size(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("size");
    
        auto result = at::size(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sizes(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sizes");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).sizes();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::slice(Tensor self, int dim=0, int start=0, int end=9223372036854775807, int step=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("slice");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 5)));
        auto start = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto end = tensor_as<int64_t>(std::move(peek(stack, 3, 5)));
        auto step = tensor_as<int64_t>(std::move(peek(stack, 4, 5)));
        auto result = at::slice(std::move(peek(stack, 0, 5)), dim, start, end, step);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto start = int64_t(node->i(Symbol::attr("start")));
      auto end = int64_t(node->i(Symbol::attr("end")));
      auto step = int64_t(node->i(Symbol::attr("step")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("slice");
    
        auto result = at::slice(std::move(peek(stack, 0, 1)), dim, start, end, step);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::slogdet(Tensor self) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("slogdet");
    
        auto result = at::slogdet(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::smm(Tensor self, Tensor mat2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("smm");
    
        auto result = at::smm(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::smooth_l1_loss(Tensor self, Tensor target, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("smooth_l1_loss");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::smooth_l1_loss(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("smooth_l1_loss");
    
        auto result = at::smooth_l1_loss(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::smooth_l1_loss_backward(Tensor grad_output, Tensor self, Tensor target, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("smooth_l1_loss_backward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::smooth_l1_loss_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), reduction);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("smooth_l1_loss_backward");
    
        auto result = at::smooth_l1_loss_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::smooth_l1_loss_forward(Tensor self, Tensor target, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("smooth_l1_loss_forward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::smooth_l1_loss_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("smooth_l1_loss_forward");
    
        auto result = at::smooth_l1_loss_forward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::soft_margin_loss(Tensor self, Tensor target, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("soft_margin_loss");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::soft_margin_loss(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("soft_margin_loss");
    
        auto result = at::soft_margin_loss(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::soft_margin_loss_backward(Tensor grad_output, Tensor self, Tensor target, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("soft_margin_loss_backward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        auto result = at::soft_margin_loss_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), std::move(peek(stack, 2, 4)), reduction);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("soft_margin_loss_backward");
    
        auto result = at::soft_margin_loss_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::soft_margin_loss_forward(Tensor self, Tensor target, int reduction) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("soft_margin_loss_forward");
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::soft_margin_loss_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("soft_margin_loss_forward");
    
        auto result = at::soft_margin_loss_forward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), reduction);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::softmax(Tensor self, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softmax");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::softmax(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softmax");
    
        auto result = at::softmax(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::softmax_backward_data(Tensor grad_output, Tensor output, int dim, Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softmax_backward_data");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto result = at::softmax_backward_data(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), dim, std::move(peek(stack, 3, 4)));
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softmax_backward_data");
    
        auto result = at::softmax_backward_data(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), dim, std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::softplus(Tensor self, Scalar beta=1, Scalar threshold=20) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softplus");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto threshold = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::softplus(std::move(peek(stack, 0, 3)), beta, threshold);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto threshold = Scalar(node->t(Symbol::attr("threshold")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softplus");
    
        auto result = at::softplus(std::move(peek(stack, 0, 1)), beta, threshold);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::softplus_backward(Tensor grad_output, Tensor self, Scalar beta, Scalar threshold, Tensor output) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softplus_backward");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 2, 5)));
        auto threshold = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto result = at::softplus_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), beta, threshold, std::move(peek(stack, 4, 5)));
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto threshold = Scalar(node->t(Symbol::attr("threshold")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softplus_backward");
    
        auto result = at::softplus_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), beta, threshold, std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::softplus_forward(Tensor self, Scalar beta, Scalar threshold) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softplus_forward");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto threshold = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::softplus_forward(std::move(peek(stack, 0, 3)), beta, threshold);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto threshold = Scalar(node->t(Symbol::attr("threshold")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softplus_forward");
    
        auto result = at::softplus_forward(std::move(peek(stack, 0, 1)), beta, threshold);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::softshrink(Tensor self, Scalar lambd=0.5) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softshrink");
        auto lambd = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::softshrink(std::move(peek(stack, 0, 2)), lambd);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto lambd = Scalar(node->t(Symbol::attr("lambd")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softshrink");
    
        auto result = at::softshrink(std::move(peek(stack, 0, 1)), lambd);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::softshrink_backward(Tensor grad_output, Tensor self, Scalar lambd) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softshrink_backward");
        auto lambd = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::softshrink_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), lambd);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto lambd = Scalar(node->t(Symbol::attr("lambd")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softshrink_backward");
    
        auto result = at::softshrink_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), lambd);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::softshrink_forward(Tensor self, Scalar lambd) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softshrink_forward");
        auto lambd = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::softshrink_forward(std::move(peek(stack, 0, 2)), lambd);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto lambd = Scalar(node->t(Symbol::attr("lambd")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("softshrink_forward");
    
        auto result = at::softshrink_forward(std::move(peek(stack, 0, 1)), lambd);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sort(Tensor self, int dim=-1, bool descending=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sort");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto descending = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::sort(std::move(peek(stack, 0, 3)), dim, descending);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto descending = bool(node->i(Symbol::attr("descending")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sort");
    
        auto result = at::sort(std::move(peek(stack, 0, 1)), dim, descending);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::split(Tensor self, int split_size, int dim=0) -> Tensor[]",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("split");
        auto split_size = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::split(std::move(peek(stack, 0, 3)), split_size, dim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto split_size = int64_t(node->i(Symbol::attr("split_size")));
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("split");
    
        auto result = at::split(std::move(peek(stack, 0, 1)), split_size, dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::split_with_sizes(Tensor self, int[] split_sizes, int dim=0) -> Tensor[]",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("split_with_sizes");
        auto split_sizes = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::split_with_sizes(std::move(peek(stack, 0, 3)), split_sizes, dim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto split_sizes = std::vector<int64_t>(node->is(Symbol::attr("split_sizes")));
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("split_with_sizes");
    
        auto result = at::split_with_sizes(std::move(peek(stack, 0, 1)), split_sizes, dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sqrt(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sqrt");
    
        auto result = at::sqrt(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::squeeze(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("squeeze");
    
        auto result = at::squeeze(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::squeeze(Tensor self, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("squeeze");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::squeeze(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("squeeze");
    
        auto result = at::squeeze(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sspaddmm(Tensor self, Tensor mat1, Tensor mat2, *, Scalar beta=1, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sspaddmm");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 4, 5)));
        auto result = at::sspaddmm(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), beta, alpha);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sspaddmm");
    
        auto result = at::sspaddmm(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), beta, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::stack(Tensor[] tensors, int dim=0) -> Tensor",
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("stack");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 0, 1)));
        auto result = at::stack(peekSlice(stack, 0, varargs_length - 1, varargs_length), dim);
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      size_t varargs_length = node->inputs().size();
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("stack");
    
        auto result = at::stack(peekSlice(stack, 0, varargs_length - 0, varargs_length), dim);
        drop(stack, varargs_length);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::std(Tensor self, bool unbiased=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("std");
        auto unbiased = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::std(std::move(peek(stack, 0, 2)), unbiased);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto unbiased = bool(node->i(Symbol::attr("unbiased")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("std");
    
        auto result = at::std(std::move(peek(stack, 0, 1)), unbiased);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::std(Tensor self, int dim, bool unbiased=True, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("std");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto unbiased = tensor_as<bool>(std::move(peek(stack, 2, 4)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::std(std::move(peek(stack, 0, 4)), dim, unbiased, keepdim);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto unbiased = bool(node->i(Symbol::attr("unbiased")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("std");
    
        auto result = at::std(std::move(peek(stack, 0, 1)), dim, unbiased, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::stft(Tensor self, int frame_length, int hop, int fft_size, bool normalized=False, bool onesided=True, Tensor? window=None, int pad_end=0) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("stft");
        auto frame_length = tensor_as<int64_t>(std::move(peek(stack, 1, 8)));
        auto hop = tensor_as<int64_t>(std::move(peek(stack, 2, 8)));
        auto fft_size = tensor_as<int64_t>(std::move(peek(stack, 3, 8)));
        auto normalized = tensor_as<bool>(std::move(peek(stack, 4, 8)));
        auto onesided = tensor_as<bool>(std::move(peek(stack, 5, 8)));
        auto pad_end = tensor_as<int64_t>(std::move(peek(stack, 7, 8)));
        auto result = at::stft(std::move(peek(stack, 0, 8)), frame_length, hop, fft_size, normalized, onesided, std::move(peek(stack, 6, 8)), pad_end);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto frame_length = int64_t(node->i(Symbol::attr("frame_length")));
      auto hop = int64_t(node->i(Symbol::attr("hop")));
      auto fft_size = int64_t(node->i(Symbol::attr("fft_size")));
      auto normalized = bool(node->i(Symbol::attr("normalized")));
      auto onesided = bool(node->i(Symbol::attr("onesided")));
      auto pad_end = int64_t(node->i(Symbol::attr("pad_end")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("stft");
    
        auto result = at::stft(std::move(peek(stack, 0, 2)), frame_length, hop, fft_size, normalized, onesided, std::move(peek(stack, 1, 2)), pad_end);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::storage_offset(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("storage_offset");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).storage_offset();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::stride(Tensor self, int dim) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("stride");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::stride(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("stride");
    
        auto result = at::stride(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::strides(Tensor self) -> int",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("strides");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).strides();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sub(Tensor self, Tensor other, *, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sub");
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::sub(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sub");
    
        auto result = at::sub(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), alpha);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sub(Tensor self, Scalar other, *, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sub");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::sub(std::move(peek(stack, 0, 3)), other, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sub");
    
        auto result = at::sub(std::move(peek(stack, 0, 1)), other, alpha);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sum(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sum");
    
        auto result = at::sum(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::sum(Tensor self, int[1] dim, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sum");
        auto dim = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::sum(std::move(peek(stack, 0, 3)), dim, keepdim);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = std::vector<int64_t>(node->is(Symbol::attr("dim")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("sum");
    
        auto result = at::sum(std::move(peek(stack, 0, 1)), dim, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::svd(Tensor self, bool some=True) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("svd");
        auto some = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::svd(std::move(peek(stack, 0, 2)), some);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto some = bool(node->i(Symbol::attr("some")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("svd");
    
        auto result = at::svd(std::move(peek(stack, 0, 1)), some);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::symeig(Tensor self, bool eigenvectors=False, bool upper=True) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("symeig");
        auto eigenvectors = tensor_as<bool>(std::move(peek(stack, 1, 3)));
        auto upper = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::symeig(std::move(peek(stack, 0, 3)), eigenvectors, upper);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto eigenvectors = bool(node->i(Symbol::attr("eigenvectors")));
      auto upper = bool(node->i(Symbol::attr("upper")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("symeig");
    
        auto result = at::symeig(std::move(peek(stack, 0, 1)), eigenvectors, upper);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::t(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("t");
    
        auto result = at::t(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::take(Tensor self, Tensor index) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("take");
    
        auto result = at::take(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::tan(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("tan");
    
        auto result = at::tan(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::tanh(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("tanh");
    
        auto result = at::tanh(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::th_add(Tensor self, Tensor other, *, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_add");
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::th_add(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_add");
    
        auto result = at::th_add(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), alpha);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::th_addmm(Tensor self, Tensor mat1, Tensor mat2, *, Scalar beta=1, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_addmm");
        auto beta = tensor_as<Scalar>(std::move(peek(stack, 3, 5)));
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 4, 5)));
        auto result = at::th_addmm(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), beta, alpha);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto beta = Scalar(node->t(Symbol::attr("beta")));
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_addmm");
    
        auto result = at::th_addmm(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), beta, alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::th_clone(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_clone");
    
        auto result = at::th_clone(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::th_div(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_div");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::th_div(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_div");
    
        auto result = at::th_div(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::th_mul(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_mul");
    
        auto result = at::th_mul(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::th_mul(Tensor self, Scalar other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_mul");
        auto other = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::th_mul(std::move(peek(stack, 0, 2)), other);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto other = Scalar(node->t(Symbol::attr("other")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_mul");
    
        auto result = at::th_mul(std::move(peek(stack, 0, 1)), other);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::th_norm(Tensor self, Scalar p=2) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_norm");
        auto p = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::th_norm(std::move(peek(stack, 0, 2)), p);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto p = Scalar(node->t(Symbol::attr("p")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_norm");
    
        auto result = at::th_norm(std::move(peek(stack, 0, 1)), p);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::th_pow(Tensor self, Scalar exponent) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_pow");
        auto exponent = tensor_as<Scalar>(std::move(peek(stack, 1, 2)));
        auto result = at::th_pow(std::move(peek(stack, 0, 2)), exponent);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto exponent = Scalar(node->t(Symbol::attr("exponent")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_pow");
    
        auto result = at::th_pow(std::move(peek(stack, 0, 1)), exponent);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::th_sub(Tensor self, Tensor other, *, Scalar alpha=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_sub");
        auto alpha = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::th_sub(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), alpha);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto alpha = Scalar(node->t(Symbol::attr("alpha")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("th_sub");
    
        auto result = at::th_sub(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), alpha);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_batch_norm(Tensor self, Tensor weight, Tensor bias, Tensor running_mean, Tensor running_var, bool training, float momentum, float eps) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_batch_norm");
        auto training = tensor_as<bool>(std::move(peek(stack, 5, 8)));
        auto momentum = tensor_as<double>(std::move(peek(stack, 6, 8)));
        auto eps = tensor_as<double>(std::move(peek(stack, 7, 8)));
        auto result = at::thnn_batch_norm(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), std::move(peek(stack, 3, 8)), std::move(peek(stack, 4, 8)), training, momentum, eps);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto training = bool(node->i(Symbol::attr("training")));
      auto momentum = double(node->f(Symbol::attr("momentum")));
      auto eps = double(node->f(Symbol::attr("eps")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_batch_norm");
    
        auto result = at::thnn_batch_norm(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), training, momentum, eps);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_batch_norm_backward(Tensor grad_output, Tensor self, Tensor? weight, Tensor? running_mean, Tensor? running_var, bool training, float eps, Tensor? save_mean, Tensor? save_std, bool[3] output_mask=[True, True, True]) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_batch_norm_backward");
        auto training = tensor_as<bool>(std::move(peek(stack, 5, 10)));
        auto eps = tensor_as<double>(std::move(peek(stack, 6, 10)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 9, 10)));
        auto result = at::thnn_batch_norm_backward(std::move(peek(stack, 0, 10)), std::move(peek(stack, 1, 10)), std::move(peek(stack, 2, 10)), std::move(peek(stack, 3, 10)), std::move(peek(stack, 4, 10)), training, eps, std::move(peek(stack, 7, 10)), std::move(peek(stack, 8, 10)), output_mask);
        drop(stack, 10);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto training = bool(node->i(Symbol::attr("training")));
      auto eps = double(node->f(Symbol::attr("eps")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_batch_norm_backward");
    
        auto result = at::thnn_batch_norm_backward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), std::move(peek(stack, 2, 7)), std::move(peek(stack, 3, 7)), std::move(peek(stack, 4, 7)), training, eps, std::move(peek(stack, 5, 7)), std::move(peek(stack, 6, 7)), output_mask);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_batch_norm_forward(Tensor self, Tensor? weight, Tensor? bias, Tensor? running_mean, Tensor? running_var, bool training, float momentum, float eps) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_batch_norm_forward");
        auto training = tensor_as<bool>(std::move(peek(stack, 5, 8)));
        auto momentum = tensor_as<double>(std::move(peek(stack, 6, 8)));
        auto eps = tensor_as<double>(std::move(peek(stack, 7, 8)));
        auto result = at::thnn_batch_norm_forward(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), std::move(peek(stack, 3, 8)), std::move(peek(stack, 4, 8)), training, momentum, eps);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto training = bool(node->i(Symbol::attr("training")));
      auto momentum = double(node->f(Symbol::attr("momentum")));
      auto eps = double(node->f(Symbol::attr("eps")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_batch_norm_forward");
    
        auto result = at::thnn_batch_norm_forward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), training, momentum, eps);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv2d(Tensor self, Tensor weight, int[2] kernel_size, Tensor bias=None, int[2] stride=1, int[2] padding=0) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv2d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        auto result = at::thnn_conv2d(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), kernel_size, std::move(peek(stack, 3, 6)), stride, padding);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv2d");
    
        auto result = at::thnn_conv2d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv2d_backward(Tensor grad_output, Tensor self, Tensor weight, int[2] kernel_size, int[2] stride, int[2] padding, Tensor finput, Tensor fgrad_input, bool[3] output_mask=[True, True, True]) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv2d_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 3, 9)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 9)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 9)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 8, 9)));
        auto result = at::thnn_conv2d_backward(std::move(peek(stack, 0, 9)), std::move(peek(stack, 1, 9)), std::move(peek(stack, 2, 9)), kernel_size, stride, padding, std::move(peek(stack, 6, 9)), std::move(peek(stack, 7, 9)), output_mask);
        drop(stack, 9);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv2d_backward");
    
        auto result = at::thnn_conv2d_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), kernel_size, stride, padding, std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), output_mask);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv2d_forward(Tensor self, Tensor weight, int[2] kernel_size, Tensor? bias, int[2] stride, int[2] padding) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv2d_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        auto result = at::thnn_conv2d_forward(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), kernel_size, std::move(peek(stack, 3, 6)), stride, padding);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv2d_forward");
    
        auto result = at::thnn_conv2d_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv3d(Tensor self, Tensor weight, int[3] kernel_size, Tensor bias=None, int[3] stride=1, int[3] padding=0) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv3d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        auto result = at::thnn_conv3d(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), kernel_size, std::move(peek(stack, 3, 6)), stride, padding);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv3d");
    
        auto result = at::thnn_conv3d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv3d_backward(Tensor grad_output, Tensor self, Tensor weight, int[3] kernel_size, int[3] stride, int[3] padding, Tensor finput, Tensor fgrad_input, bool[3] output_mask=[True, True, True]) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv3d_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 3, 9)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 9)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 9)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 8, 9)));
        auto result = at::thnn_conv3d_backward(std::move(peek(stack, 0, 9)), std::move(peek(stack, 1, 9)), std::move(peek(stack, 2, 9)), kernel_size, stride, padding, std::move(peek(stack, 6, 9)), std::move(peek(stack, 7, 9)), output_mask);
        drop(stack, 9);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv3d_backward");
    
        auto result = at::thnn_conv3d_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), kernel_size, stride, padding, std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), output_mask);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv3d_forward(Tensor self, Tensor weight, int[3] kernel_size, Tensor? bias, int[3] stride, int[3] padding) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv3d_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 6)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 6)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 6)));
        auto result = at::thnn_conv3d_forward(std::move(peek(stack, 0, 6)), std::move(peek(stack, 1, 6)), kernel_size, std::move(peek(stack, 3, 6)), stride, padding);
        drop(stack, 6);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv3d_forward");
    
        auto result = at::thnn_conv3d_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_depthwise2d(Tensor self, Tensor weight, int[2] kernel_size, Tensor bias=None, int[2] stride=1, int[2] padding=0, int[2] dilation=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_depthwise2d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 7)));
        auto result = at::thnn_conv_depthwise2d(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), kernel_size, std::move(peek(stack, 3, 7)), stride, padding, dilation);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_depthwise2d");
    
        auto result = at::thnn_conv_depthwise2d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_depthwise2d_backward(Tensor grad_output, Tensor self, Tensor weight, int[2] kernel_size, int[2] stride, int[2] padding, int[2] dilation, bool[2] output_mask=[True, True]) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_depthwise2d_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 3, 8)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 8)));
        auto output_mask = tensor_as<std::array<bool,2>>(std::move(peek(stack, 7, 8)));
        auto result = at::thnn_conv_depthwise2d_backward(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), kernel_size, stride, padding, dilation, output_mask);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto output_mask = as_bool_array<2>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_depthwise2d_backward");
    
        auto result = at::thnn_conv_depthwise2d_backward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), kernel_size, stride, padding, dilation, output_mask);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_depthwise2d_forward(Tensor self, Tensor weight, int[2] kernel_size, Tensor? bias, int[2] stride, int[2] padding, int[2] dilation) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_depthwise2d_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 7)));
        auto result = at::thnn_conv_depthwise2d_forward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), kernel_size, std::move(peek(stack, 3, 7)), stride, padding, dilation);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_depthwise2d_forward");
    
        auto result = at::thnn_conv_depthwise2d_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_dilated2d(Tensor self, Tensor weight, int[2] kernel_size, Tensor bias=None, int[2] stride=1, int[2] padding=0, int[2] dilation=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated2d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 7)));
        auto result = at::thnn_conv_dilated2d(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), kernel_size, std::move(peek(stack, 3, 7)), stride, padding, dilation);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated2d");
    
        auto result = at::thnn_conv_dilated2d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_dilated2d_backward(Tensor grad_output, Tensor self, Tensor weight, int[2] kernel_size, int[2] stride, int[2] padding, int[2] dilation, Tensor columns, Tensor ones, bool[3] output_mask=[True, True, True]) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated2d_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 3, 10)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 10)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 10)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 10)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 9, 10)));
        auto result = at::thnn_conv_dilated2d_backward(std::move(peek(stack, 0, 10)), std::move(peek(stack, 1, 10)), std::move(peek(stack, 2, 10)), kernel_size, stride, padding, dilation, std::move(peek(stack, 7, 10)), std::move(peek(stack, 8, 10)), output_mask);
        drop(stack, 10);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated2d_backward");
    
        auto result = at::thnn_conv_dilated2d_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), kernel_size, stride, padding, dilation, std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), output_mask);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_dilated2d_forward(Tensor self, Tensor weight, int[2] kernel_size, Tensor? bias, int[2] stride, int[2] padding, int[2] dilation) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated2d_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 7)));
        auto result = at::thnn_conv_dilated2d_forward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), kernel_size, std::move(peek(stack, 3, 7)), stride, padding, dilation);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated2d_forward");
    
        auto result = at::thnn_conv_dilated2d_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_dilated3d(Tensor self, Tensor weight, int[3] kernel_size, Tensor bias=None, int[3] stride=1, int[3] padding=0, int[3] dilation=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated3d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 7)));
        auto result = at::thnn_conv_dilated3d(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), kernel_size, std::move(peek(stack, 3, 7)), stride, padding, dilation);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated3d");
    
        auto result = at::thnn_conv_dilated3d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_dilated3d_backward(Tensor grad_output, Tensor self, Tensor weight, int[3] kernel_size, int[3] stride, int[3] padding, int[3] dilation, Tensor columns, Tensor ones, bool[3] output_mask=[True, True, True]) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated3d_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 3, 10)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 10)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 10)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 10)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 9, 10)));
        auto result = at::thnn_conv_dilated3d_backward(std::move(peek(stack, 0, 10)), std::move(peek(stack, 1, 10)), std::move(peek(stack, 2, 10)), kernel_size, stride, padding, dilation, std::move(peek(stack, 7, 10)), std::move(peek(stack, 8, 10)), output_mask);
        drop(stack, 10);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated3d_backward");
    
        auto result = at::thnn_conv_dilated3d_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), kernel_size, stride, padding, dilation, std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), output_mask);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_dilated3d_forward(Tensor self, Tensor weight, int[3] kernel_size, Tensor? bias, int[3] stride, int[3] padding, int[3] dilation) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated3d_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 7)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 7)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 7)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 6, 7)));
        auto result = at::thnn_conv_dilated3d_forward(std::move(peek(stack, 0, 7)), std::move(peek(stack, 1, 7)), kernel_size, std::move(peek(stack, 3, 7)), stride, padding, dilation);
        drop(stack, 7);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_dilated3d_forward");
    
        auto result = at::thnn_conv_dilated3d_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_transpose2d(Tensor self, Tensor weight, int[2] kernel_size, Tensor bias=None, int[2] stride=1, int[2] padding=0, int[2] output_padding=0, int[2] dilation=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose2d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 8)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 6, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 7, 8)));
        auto result = at::thnn_conv_transpose2d(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), kernel_size, std::move(peek(stack, 3, 8)), stride, padding, output_padding, dilation);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose2d");
    
        auto result = at::thnn_conv_transpose2d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding, output_padding, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_transpose2d_backward(Tensor grad_output, Tensor self, Tensor weight, int[2] kernel_size, int[2] stride, int[2] padding, int[2] output_padding, int[2] dilation, Tensor columns, Tensor ones, bool[3] output_mask=[True, True, True]) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose2d_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 3, 11)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 11)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 11)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 6, 11)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 7, 11)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 10, 11)));
        auto result = at::thnn_conv_transpose2d_backward(std::move(peek(stack, 0, 11)), std::move(peek(stack, 1, 11)), std::move(peek(stack, 2, 11)), kernel_size, stride, padding, output_padding, dilation, std::move(peek(stack, 8, 11)), std::move(peek(stack, 9, 11)), output_mask);
        drop(stack, 11);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose2d_backward");
    
        auto result = at::thnn_conv_transpose2d_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), kernel_size, stride, padding, output_padding, dilation, std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), output_mask);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_transpose2d_forward(Tensor self, Tensor weight, int[2] kernel_size, Tensor? bias, int[2] stride, int[2] padding, int[2] output_padding, int[2] dilation) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose2d_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 8)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 6, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 7, 8)));
        auto result = at::thnn_conv_transpose2d_forward(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), kernel_size, std::move(peek(stack, 3, 8)), stride, padding, output_padding, dilation);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose2d_forward");
    
        auto result = at::thnn_conv_transpose2d_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding, output_padding, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_transpose3d(Tensor self, Tensor weight, int[3] kernel_size, Tensor bias=None, int[3] stride=1, int[3] padding=0, int[3] output_padding=0, int[3] dilation=1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose3d");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 8)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 6, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 7, 8)));
        auto result = at::thnn_conv_transpose3d(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), kernel_size, std::move(peek(stack, 3, 8)), stride, padding, output_padding, dilation);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose3d");
    
        auto result = at::thnn_conv_transpose3d(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding, output_padding, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_transpose3d_backward(Tensor grad_output, Tensor self, Tensor weight, int[3] kernel_size, int[3] stride, int[3] padding, int[3] output_padding, int[3] dilation, Tensor finput, Tensor fgrad_input, bool[3] output_mask=[True, True, True]) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose3d_backward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 3, 11)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 11)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 11)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 6, 11)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 7, 11)));
        auto output_mask = tensor_as<std::array<bool,3>>(std::move(peek(stack, 10, 11)));
        auto result = at::thnn_conv_transpose3d_backward(std::move(peek(stack, 0, 11)), std::move(peek(stack, 1, 11)), std::move(peek(stack, 2, 11)), kernel_size, stride, padding, output_padding, dilation, std::move(peek(stack, 8, 11)), std::move(peek(stack, 9, 11)), output_mask);
        drop(stack, 11);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      auto output_mask = as_bool_array<3>(node->is(Symbol::attr("output_mask")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose3d_backward");
    
        auto result = at::thnn_conv_transpose3d_backward(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), std::move(peek(stack, 2, 5)), kernel_size, stride, padding, output_padding, dilation, std::move(peek(stack, 3, 5)), std::move(peek(stack, 4, 5)), output_mask);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::thnn_conv_transpose3d_forward(Tensor self, Tensor weight, int[3] kernel_size, Tensor? bias, int[3] stride, int[3] padding, int[3] output_padding, int[3] dilation) -> (Tensor, Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose3d_forward");
        auto kernel_size = tensor_as<IntList>(std::move(peek(stack, 2, 8)));
        auto stride = tensor_as<IntList>(std::move(peek(stack, 4, 8)));
        auto padding = tensor_as<IntList>(std::move(peek(stack, 5, 8)));
        auto output_padding = tensor_as<IntList>(std::move(peek(stack, 6, 8)));
        auto dilation = tensor_as<IntList>(std::move(peek(stack, 7, 8)));
        auto result = at::thnn_conv_transpose3d_forward(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), kernel_size, std::move(peek(stack, 3, 8)), stride, padding, output_padding, dilation);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto kernel_size = std::vector<int64_t>(node->is(Symbol::attr("kernel_size")));
      auto stride = std::vector<int64_t>(node->is(Symbol::attr("stride")));
      auto padding = std::vector<int64_t>(node->is(Symbol::attr("padding")));
      auto output_padding = std::vector<int64_t>(node->is(Symbol::attr("output_padding")));
      auto dilation = std::vector<int64_t>(node->is(Symbol::attr("dilation")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("thnn_conv_transpose3d_forward");
    
        auto result = at::thnn_conv_transpose3d_forward(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), kernel_size, std::move(peek(stack, 2, 3)), stride, padding, output_padding, dilation);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::threshold(Tensor self, Scalar threshold, Scalar value) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("threshold");
        auto threshold = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto value = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::threshold(std::move(peek(stack, 0, 3)), threshold, value);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto threshold = Scalar(node->t(Symbol::attr("threshold")));
      auto value = Scalar(node->t(Symbol::attr("value")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("threshold");
    
        auto result = at::threshold(std::move(peek(stack, 0, 1)), threshold, value);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::threshold_backward(Tensor grad_output, Tensor self, Scalar threshold, Scalar value) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("threshold_backward");
        auto threshold = tensor_as<Scalar>(std::move(peek(stack, 2, 4)));
        auto value = tensor_as<Scalar>(std::move(peek(stack, 3, 4)));
        auto result = at::threshold_backward(std::move(peek(stack, 0, 4)), std::move(peek(stack, 1, 4)), threshold, value);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto threshold = Scalar(node->t(Symbol::attr("threshold")));
      auto value = Scalar(node->t(Symbol::attr("value")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("threshold_backward");
    
        auto result = at::threshold_backward(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), threshold, value);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::threshold_forward(Tensor self, Scalar threshold, Scalar value) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("threshold_forward");
        auto threshold = tensor_as<Scalar>(std::move(peek(stack, 1, 3)));
        auto value = tensor_as<Scalar>(std::move(peek(stack, 2, 3)));
        auto result = at::threshold_forward(std::move(peek(stack, 0, 3)), threshold, value);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto threshold = Scalar(node->t(Symbol::attr("threshold")));
      auto value = Scalar(node->t(Symbol::attr("value")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("threshold_forward");
    
        auto result = at::threshold_forward(std::move(peek(stack, 0, 1)), threshold, value);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::to_dense(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("to_dense");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).to_dense();
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::topk(Tensor self, int k, int dim=-1, bool largest=True, bool sorted=True) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("topk");
        auto k = tensor_as<int64_t>(std::move(peek(stack, 1, 5)));
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 2, 5)));
        auto largest = tensor_as<bool>(std::move(peek(stack, 3, 5)));
        auto sorted = tensor_as<bool>(std::move(peek(stack, 4, 5)));
        auto result = at::topk(std::move(peek(stack, 0, 5)), k, dim, largest, sorted);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto k = int64_t(node->i(Symbol::attr("k")));
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto largest = bool(node->i(Symbol::attr("largest")));
      auto sorted = bool(node->i(Symbol::attr("sorted")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("topk");
    
        auto result = at::topk(std::move(peek(stack, 0, 1)), k, dim, largest, sorted);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::trace(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("trace");
    
        auto result = at::trace(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::transpose(Tensor self, int dim0, int dim1) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("transpose");
        auto dim0 = tensor_as<int64_t>(std::move(peek(stack, 1, 3)));
        auto dim1 = tensor_as<int64_t>(std::move(peek(stack, 2, 3)));
        auto result = at::transpose(std::move(peek(stack, 0, 3)), dim0, dim1);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim0 = int64_t(node->i(Symbol::attr("dim0")));
      auto dim1 = int64_t(node->i(Symbol::attr("dim1")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("transpose");
    
        auto result = at::transpose(std::move(peek(stack, 0, 1)), dim0, dim1);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::tril(Tensor self, int diagonal=0) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("tril");
        auto diagonal = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::tril(std::move(peek(stack, 0, 2)), diagonal);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto diagonal = int64_t(node->i(Symbol::attr("diagonal")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("tril");
    
        auto result = at::tril(std::move(peek(stack, 0, 1)), diagonal);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::triplet_margin_loss(Tensor anchor, Tensor positive, Tensor negative, float margin=1.0, float p=2, float eps=1e-06, bool swap=False, int reduction=ElementwiseMean) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("triplet_margin_loss");
        auto margin = tensor_as<double>(std::move(peek(stack, 3, 8)));
        auto p = tensor_as<double>(std::move(peek(stack, 4, 8)));
        auto eps = tensor_as<double>(std::move(peek(stack, 5, 8)));
        auto swap = tensor_as<bool>(std::move(peek(stack, 6, 8)));
        auto reduction = tensor_as<int64_t>(std::move(peek(stack, 7, 8)));
        auto result = at::triplet_margin_loss(std::move(peek(stack, 0, 8)), std::move(peek(stack, 1, 8)), std::move(peek(stack, 2, 8)), margin, p, eps, swap, reduction);
        drop(stack, 8);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto margin = double(node->f(Symbol::attr("margin")));
      auto p = double(node->f(Symbol::attr("p")));
      auto eps = double(node->f(Symbol::attr("eps")));
      auto swap = bool(node->i(Symbol::attr("swap")));
      auto reduction = int64_t(node->i(Symbol::attr("reduction")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("triplet_margin_loss");
    
        auto result = at::triplet_margin_loss(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)), margin, p, eps, swap, reduction);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::triu(Tensor self, int diagonal=0) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("triu");
        auto diagonal = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::triu(std::move(peek(stack, 0, 2)), diagonal);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto diagonal = int64_t(node->i(Symbol::attr("diagonal")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("triu");
    
        auto result = at::triu(std::move(peek(stack, 0, 1)), diagonal);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::trtrs(Tensor self, Tensor A, bool upper=True, bool transpose=False, bool unitriangular=False) -> (Tensor, Tensor)",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("trtrs");
        auto upper = tensor_as<bool>(std::move(peek(stack, 2, 5)));
        auto transpose = tensor_as<bool>(std::move(peek(stack, 3, 5)));
        auto unitriangular = tensor_as<bool>(std::move(peek(stack, 4, 5)));
        auto result = at::trtrs(std::move(peek(stack, 0, 5)), std::move(peek(stack, 1, 5)), upper, transpose, unitriangular);
        drop(stack, 5);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto upper = bool(node->i(Symbol::attr("upper")));
      auto transpose = bool(node->i(Symbol::attr("transpose")));
      auto unitriangular = bool(node->i(Symbol::attr("unitriangular")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("trtrs");
    
        auto result = at::trtrs(std::move(peek(stack, 0, 2)), std::move(peek(stack, 1, 2)), upper, transpose, unitriangular);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::trunc(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("trunc");
    
        auto result = at::trunc(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::type_as(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("type_as");
    
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).type_as(std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::unbind(Tensor self, int dim=0) -> Tensor[]",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("unbind");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::unbind(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("unbind");
    
        auto result = at::unbind(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::unfold(Tensor self, int dimension, int size, int step) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("unfold");
        auto dimension = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto size = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto step = tensor_as<int64_t>(std::move(peek(stack, 3, 4)));
        DeviceGuard device_guard(deviceForInputs(stack, 4));
        auto result = (std::move(peek(stack, 0, 4))).unfold(dimension, size, step);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dimension = int64_t(node->i(Symbol::attr("dimension")));
      auto size = int64_t(node->i(Symbol::attr("size")));
      auto step = int64_t(node->i(Symbol::attr("step")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("unfold");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).unfold(dimension, size, step);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::unsqueeze(Tensor self, int dim) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("unsqueeze");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 2)));
        auto result = at::unsqueeze(std::move(peek(stack, 0, 2)), dim);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("unsqueeze");
    
        auto result = at::unsqueeze(std::move(peek(stack, 0, 1)), dim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_bilinear2d(Tensor self, int[2] output_size, bool align_corners) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_bilinear2d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto align_corners = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::upsample_bilinear2d(std::move(peek(stack, 0, 3)), output_size, align_corners);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto align_corners = bool(node->i(Symbol::attr("align_corners")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_bilinear2d");
    
        auto result = at::upsample_bilinear2d(std::move(peek(stack, 0, 1)), output_size, align_corners);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_bilinear2d_backward(Tensor grad_output, int[2] output_size, int[4] input_size, bool align_corners) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_bilinear2d_backward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 4)));
        auto input_size = tensor_as<IntList>(std::move(peek(stack, 2, 4)));
        auto align_corners = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::upsample_bilinear2d_backward(std::move(peek(stack, 0, 4)), output_size, input_size, align_corners);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto input_size = std::vector<int64_t>(node->is(Symbol::attr("input_size")));
      auto align_corners = bool(node->i(Symbol::attr("align_corners")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_bilinear2d_backward");
    
        auto result = at::upsample_bilinear2d_backward(std::move(peek(stack, 0, 1)), output_size, input_size, align_corners);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_bilinear2d_forward(Tensor self, int[2] output_size, bool align_corners) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_bilinear2d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto align_corners = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::upsample_bilinear2d_forward(std::move(peek(stack, 0, 3)), output_size, align_corners);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto align_corners = bool(node->i(Symbol::attr("align_corners")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_bilinear2d_forward");
    
        auto result = at::upsample_bilinear2d_forward(std::move(peek(stack, 0, 1)), output_size, align_corners);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_linear1d(Tensor self, int[1] output_size, bool align_corners) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_linear1d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto align_corners = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::upsample_linear1d(std::move(peek(stack, 0, 3)), output_size, align_corners);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto align_corners = bool(node->i(Symbol::attr("align_corners")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_linear1d");
    
        auto result = at::upsample_linear1d(std::move(peek(stack, 0, 1)), output_size, align_corners);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_linear1d_backward(Tensor grad_output, int[1] output_size, int[3] input_size, bool align_corners) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_linear1d_backward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 4)));
        auto input_size = tensor_as<IntList>(std::move(peek(stack, 2, 4)));
        auto align_corners = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::upsample_linear1d_backward(std::move(peek(stack, 0, 4)), output_size, input_size, align_corners);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto input_size = std::vector<int64_t>(node->is(Symbol::attr("input_size")));
      auto align_corners = bool(node->i(Symbol::attr("align_corners")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_linear1d_backward");
    
        auto result = at::upsample_linear1d_backward(std::move(peek(stack, 0, 1)), output_size, input_size, align_corners);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_linear1d_forward(Tensor self, int[1] output_size, bool align_corners) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_linear1d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto align_corners = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::upsample_linear1d_forward(std::move(peek(stack, 0, 3)), output_size, align_corners);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto align_corners = bool(node->i(Symbol::attr("align_corners")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_linear1d_forward");
    
        auto result = at::upsample_linear1d_forward(std::move(peek(stack, 0, 1)), output_size, align_corners);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_nearest1d(Tensor self, int[1] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest1d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::upsample_nearest1d(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest1d");
    
        auto result = at::upsample_nearest1d(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_nearest1d_backward(Tensor grad_output, int[1] output_size, int[3] input_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest1d_backward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto input_size = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::upsample_nearest1d_backward(std::move(peek(stack, 0, 3)), output_size, input_size);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto input_size = std::vector<int64_t>(node->is(Symbol::attr("input_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest1d_backward");
    
        auto result = at::upsample_nearest1d_backward(std::move(peek(stack, 0, 1)), output_size, input_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_nearest1d_forward(Tensor self, int[1] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest1d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::upsample_nearest1d_forward(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest1d_forward");
    
        auto result = at::upsample_nearest1d_forward(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_nearest2d(Tensor self, int[2] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest2d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::upsample_nearest2d(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest2d");
    
        auto result = at::upsample_nearest2d(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_nearest2d_backward(Tensor grad_output, int[2] output_size, int[4] input_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest2d_backward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto input_size = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::upsample_nearest2d_backward(std::move(peek(stack, 0, 3)), output_size, input_size);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto input_size = std::vector<int64_t>(node->is(Symbol::attr("input_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest2d_backward");
    
        auto result = at::upsample_nearest2d_backward(std::move(peek(stack, 0, 1)), output_size, input_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_nearest2d_forward(Tensor self, int[2] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest2d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::upsample_nearest2d_forward(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest2d_forward");
    
        auto result = at::upsample_nearest2d_forward(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_nearest3d(Tensor self, int[3] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest3d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::upsample_nearest3d(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest3d");
    
        auto result = at::upsample_nearest3d(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_nearest3d_backward(Tensor grad_output, int[3] output_size, int[5] input_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest3d_backward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto input_size = tensor_as<IntList>(std::move(peek(stack, 2, 3)));
        auto result = at::upsample_nearest3d_backward(std::move(peek(stack, 0, 3)), output_size, input_size);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto input_size = std::vector<int64_t>(node->is(Symbol::attr("input_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest3d_backward");
    
        auto result = at::upsample_nearest3d_backward(std::move(peek(stack, 0, 1)), output_size, input_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_nearest3d_forward(Tensor self, int[3] output_size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest3d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        auto result = at::upsample_nearest3d_forward(std::move(peek(stack, 0, 2)), output_size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_nearest3d_forward");
    
        auto result = at::upsample_nearest3d_forward(std::move(peek(stack, 0, 1)), output_size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_trilinear3d(Tensor self, int[3] output_size, bool align_corners) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_trilinear3d");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto align_corners = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::upsample_trilinear3d(std::move(peek(stack, 0, 3)), output_size, align_corners);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto align_corners = bool(node->i(Symbol::attr("align_corners")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_trilinear3d");
    
        auto result = at::upsample_trilinear3d(std::move(peek(stack, 0, 1)), output_size, align_corners);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_trilinear3d_backward(Tensor grad_output, int[3] output_size, int[5] input_size, bool align_corners) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_trilinear3d_backward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 4)));
        auto input_size = tensor_as<IntList>(std::move(peek(stack, 2, 4)));
        auto align_corners = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::upsample_trilinear3d_backward(std::move(peek(stack, 0, 4)), output_size, input_size, align_corners);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto input_size = std::vector<int64_t>(node->is(Symbol::attr("input_size")));
      auto align_corners = bool(node->i(Symbol::attr("align_corners")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_trilinear3d_backward");
    
        auto result = at::upsample_trilinear3d_backward(std::move(peek(stack, 0, 1)), output_size, input_size, align_corners);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::upsample_trilinear3d_forward(Tensor self, int[3] output_size, bool align_corners) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_trilinear3d_forward");
        auto output_size = tensor_as<IntList>(std::move(peek(stack, 1, 3)));
        auto align_corners = tensor_as<bool>(std::move(peek(stack, 2, 3)));
        auto result = at::upsample_trilinear3d_forward(std::move(peek(stack, 0, 3)), output_size, align_corners);
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto output_size = std::vector<int64_t>(node->is(Symbol::attr("output_size")));
      auto align_corners = bool(node->i(Symbol::attr("align_corners")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("upsample_trilinear3d_forward");
    
        auto result = at::upsample_trilinear3d_forward(std::move(peek(stack, 0, 1)), output_size, align_corners);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::var(Tensor self, bool unbiased=True) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("var");
        auto unbiased = tensor_as<bool>(std::move(peek(stack, 1, 2)));
        auto result = at::var(std::move(peek(stack, 0, 2)), unbiased);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto unbiased = bool(node->i(Symbol::attr("unbiased")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("var");
    
        auto result = at::var(std::move(peek(stack, 0, 1)), unbiased);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::var(Tensor self, int dim, bool unbiased=True, bool keepdim=False) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("var");
        auto dim = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto unbiased = tensor_as<bool>(std::move(peek(stack, 2, 4)));
        auto keepdim = tensor_as<bool>(std::move(peek(stack, 3, 4)));
        auto result = at::var(std::move(peek(stack, 0, 4)), dim, unbiased, keepdim);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dim = int64_t(node->i(Symbol::attr("dim")));
      auto unbiased = bool(node->i(Symbol::attr("unbiased")));
      auto keepdim = bool(node->i(Symbol::attr("keepdim")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("var");
    
        auto result = at::var(std::move(peek(stack, 0, 1)), dim, unbiased, keepdim);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::view(Tensor self, int[] size) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("view");
        auto size = tensor_as<IntList>(std::move(peek(stack, 1, 2)));
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).view(size);
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("view");
    
        DeviceGuard device_guard(deviceForInputs(stack, 1));
        auto result = (std::move(peek(stack, 0, 1))).view(size);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::view_as(Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("view_as");
    
        DeviceGuard device_guard(deviceForInputs(stack, 2));
        auto result = (std::move(peek(stack, 0, 2))).view_as(std::move(peek(stack, 1, 2)));
        drop(stack, 2);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::where(Tensor condition, Tensor self, Tensor other) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("where");
    
        auto result = at::where(std::move(peek(stack, 0, 3)), std::move(peek(stack, 1, 3)), std::move(peek(stack, 2, 3)));
        drop(stack, 3);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::zeros(int[] size, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("zeros");
        auto size = tensor_as<IntList>(std::move(peek(stack, 0, 4)));
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::zeros(size, options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto size = std::vector<int64_t>(node->is(Symbol::attr("size")));
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("zeros");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::zeros(size, options);
        drop(stack, 0);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::zeros_like(Tensor self) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("zeros_like");
    
        auto result = at::zeros_like(std::move(peek(stack, 0, 1)));
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
Operator(
    "aten::zeros_like(Tensor self, *, ScalarType dtype=float, Layout layout=strided, Device device=[cpu, -1]) -> Tensor",
    [](Node *node) {
    
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("zeros_like");
        auto dtype = tensor_as<int64_t>(std::move(peek(stack, 1, 4)));
        auto layout = tensor_as<int64_t>(std::move(peek(stack, 2, 4)));
        auto device = tensor_as<IntList>(std::move(peek(stack, 3, 4)));
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::zeros_like(std::move(peek(stack, 0, 4)), options);
        drop(stack, 4);
        pack(stack, std::move(result));
        return 0;
      });
    }
    ,
    [](Node *node) {
      auto dtype = int64_t(node->i(Symbol::attr("dtype")));
      auto layout = int64_t(node->i(Symbol::attr("layout")));
      auto device = std::vector<int64_t>(node->is(Symbol::attr("device")));
      return Operation([=](Stack & stack) {
        autograd::profiler::RecordFunction record("zeros_like");
    
        const auto device_index = static_cast<int32_t>(device[1]);
        const auto options = TensorOptions()
                .dtype(static_cast<at::ScalarType>(dtype))
                .layout(static_cast<at::Layout>(layout))
                .device({static_cast<at::Device::Type>(device[0]), device_index});
        auto result = torch::zeros_like(std::move(peek(stack, 0, 1)), options);
        drop(stack, 1);
        pack(stack, std::move(result));
        return 0;
      });
    }
),
});

} // anon namespace


}} // namespace torch::jit
